from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.forks.byzantium import (
    ByzantiumVM,
    get_uncle_reward,
)
from eth.vm.state import BaseState

from .blocks import ConstantinopleBlock
from .constants import EIP1234_BLOCK_REWARD
from .headers import (
    compute_constantinople_difficulty,
    configure_constantinople_header,
    create_constantinople_header_from_parent,
)
from .state import ConstantinopleState


class ConstantinopleVM(ByzantiumVM):
    # fork name
    fork = "constantinople"

    # classes
    block_class: Type[BaseBlock] = ConstantinopleBlock
    _state_class: Type[BaseState] = ConstantinopleState

    # Methods
    create_header_from_parent = staticmethod(create_constantinople_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_constantinople_difficulty)  # type: ignore
    configure_header = configure_constantinople_header
    get_uncle_reward = staticmethod(get_uncle_reward(EIP1234_BLOCK_REWARD))

    @staticmethod
    def get_block_reward() -> int:
        return EIP1234_BLOCK_REWARD



from typing import (
    Tuple,
)

from eth_utils import (
    ValidationError,
)
from py_ecc import (
    optimized_bn128 as bn128,
)
from py_ecc.optimized_bn128 import (
    FQ2,
    FQP,
)


def validate_point(x: int, y: int) -> Tuple[bn128.FQ, bn128.FQ, bn128.FQ]:
    FQ = bn128.FQ

    if x >= bn128.field_modulus:
        raise ValidationError("Point x value is greater than field modulus")
    elif y >= bn128.field_modulus:
        raise ValidationError("Point y value is greater than field modulus")

    if (x, y) != (0, 0):
        p1 = (FQ(x), FQ(y), FQ(1))
        if not bn128.is_on_curve(p1, bn128.b):
            raise ValidationError("Point is not on the curve")
    else:
        p1 = (FQ(1), FQ(1), FQ(0))

    return p1


def FQP_point_to_FQ2_point(pt: Tuple[FQP, FQP, FQP]) -> Tuple[FQ2, FQ2, FQ2]:
    """
    Transform FQP to FQ2 for type hinting.
    """
    return (
        FQ2(pt[0].coeffs),
        FQ2(pt[1].coeffs),
        FQ2(pt[2].coeffs),
    )



from eth.exceptions import (
    ReservedBytesInCode,
)
from eth.vm.forks.berlin.computation import (
    BerlinComputation,
)

from ..london.constants import (
    EIP3541_RESERVED_STARTING_BYTE,
)
from .opcodes import (
    LONDON_OPCODES,
)


class LondonComputation(BerlinComputation):
    """
    A class for all execution *message* computations in the ``London`` fork.
    Inherits from :class:`~eth.vm.forks.berlin.BerlinComputation`
    """

    opcodes = LONDON_OPCODES

    @classmethod
    def validate_contract_code(cls, contract_code: bytes) -> None:
        super().validate_contract_code(contract_code)

        if contract_code[:1] == EIP3541_RESERVED_STARTING_BYTE:
            raise ReservedBytesInCode(
                "Contract code begins with EIP3541 reserved byte '0xEF'."
            )



from typing import (
    Any,
)
from eth_typing import Address

import rlp
from rlp.sedes import (
    big_endian_int,
)
from web3 import Web3
from web3.types import (
    BlockIdentifier,
)

from eth.abc import (
    AccountAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
    EMPTY_SHA3,
)
from eth.db.journal import JournalDB

from .sedes import (
    hash32,
    trie_root,
)


class Account(rlp.Serializable, AccountAPI):
    """
    RLP object for accounts.
    """

    fields = [
        ("nonce", big_endian_int),
        ("balance", big_endian_int),
        ("storage_root", trie_root),
        ("code_hash", hash32),
    ]

    def __init__(
        self,
        nonce: int = 0,
        balance: int = 0,
        storage_root: bytes = BLANK_ROOT_HASH,
        code_hash: bytes = EMPTY_SHA3,
        **kwargs: Any,
    ) -> None:
        super().__init__(nonce, balance, storage_root, code_hash, **kwargs)

    def __repr__(self) -> str:
        return (
            f"Account(nonce={self.nonce}, balance={self.balance}, "
            f"storage_root=0x{self.storage_root.hex()}, "
            f"code_hash=0x{self.code_hash.hex()})"
        )

class LazyAccount(AccountAPI):
    _nonce: big_endian_int = None
    _balance: big_endian_int = None
    _storage_root: trie_root = None
    _code_hash: hash32 = None

    _web3: Web3
    _block_identifier: BlockIdentifier
    _address: Address

    def __init__(
            self,
            web3: Web3,
            block_identifier: BlockIdentifier,
            address: Address,
    ) -> None:
        self._web3 = web3
        self._block_identifier = block_identifier
        self._address = address

    @property
    def nonce(self) -> big_endian_int:
        if self._nonce is None:
            self._nonce = self._web3.eth.get_transaction_count(
                self._address,
                block_identifier=self._block_identifier,
            )
        return self._nonce

    @property
    def balance(self) -> big_endian_int:
        if self._balance is None:
            self._balance = self._web3.eth.get_balance(
                self._address,
                block_identifier=self._block_identifier,
            )
        return self._balance

    @property
    def storage_root(self) -> trie_root:
        raise NotImplementedError("Too lazy to implement")
    
    @property
    def code_hash(self) -> hash32:
        raise NotImplementedError("Too lazy to implement")

    def copy(self, **kwargs) -> AccountAPI:
        ret = LazyAccount(
            web3=self._web3,
            block_identifier=self._block_identifier,
            address=self._address,
        )
        ret._nonce = kwargs.get('nonce', self._nonce)
        ret._balance = kwargs.get('balance', self._balance)

        return ret

    def __str__(self) -> str:
        return f'LazyAccount(address={self._address.hex()}; _nonce={self._nonce}; _balance={self._balance})'



from rlp.sedes import (
    CountableList,
)

from eth.rlp.headers import (
    BlockHeader,
)
from eth.vm.forks.spurious_dragon.blocks import (
    SpuriousDragonBlock,
)

from .transactions import (
    ByzantiumTransaction,
)


class ByzantiumBlock(SpuriousDragonBlock):
    transaction_builder = ByzantiumTransaction
    fields = [
        ("header", BlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(BlockHeader)),
    ]



import copy

from eth_utils.toolz import (
    merge,
)

from eth.constants import (
    GAS_CREATE,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.homestead.opcodes import (
    HOMESTEAD_OPCODES,
)
from eth.vm.forks.tangerine_whistle import (
    constants,
)
from eth.vm.logic import (
    call,
    context,
    storage,
    system,
)
from eth.vm.opcode import (
    as_opcode,
)

UPDATED_OPCODES = {
    opcode_values.EXTCODESIZE: as_opcode(
        logic_fn=context.extcodesize,
        mnemonic=mnemonics.EXTCODESIZE,
        gas_cost=constants.GAS_EXTCODE_EIP150,
    ),
    opcode_values.EXTCODECOPY: as_opcode(
        logic_fn=context.extcodecopy,
        mnemonic=mnemonics.EXTCODECOPY,
        gas_cost=constants.GAS_EXTCODE_EIP150,
    ),
    opcode_values.BALANCE: as_opcode(
        logic_fn=context.balance,
        mnemonic=mnemonics.BALANCE,
        gas_cost=constants.GAS_BALANCE_EIP150,
    ),
    opcode_values.SLOAD: as_opcode(
        logic_fn=storage.sload,
        mnemonic=mnemonics.SLOAD,
        gas_cost=constants.GAS_SLOAD_EIP150,
    ),
    opcode_values.SELFDESTRUCT: as_opcode(
        logic_fn=system.selfdestruct_eip150,
        mnemonic=mnemonics.SELFDESTRUCT,
        gas_cost=constants.GAS_SELFDESTRUCT_EIP150,
    ),
    opcode_values.CREATE: system.CreateEIP150.configure(
        __name__="opcode:CREATE",
        mnemonic=mnemonics.CREATE,
        gas_cost=GAS_CREATE,
    )(),
    opcode_values.CALL: call.CallEIP150.configure(
        __name__="opcode:CALL",
        mnemonic=mnemonics.CALL,
        gas_cost=constants.GAS_CALL_EIP150,
    )(),
    opcode_values.CALLCODE: call.CallCodeEIP150.configure(
        __name__="opcode:CALLCODE",
        mnemonic=mnemonics.CALLCODE,
        gas_cost=constants.GAS_CALL_EIP150,
    )(),
    opcode_values.DELEGATECALL: call.DelegateCallEIP150.configure(
        __name__="opcode:DELEGATECALL",
        mnemonic=mnemonics.DELEGATECALL,
        gas_cost=constants.GAS_CALL_EIP150,
    )(),
}


TANGERINE_WHISTLE_OPCODES = merge(
    copy.deepcopy(HOMESTEAD_OPCODES),
    UPDATED_OPCODES,
)



import copy
from typing import (
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.abc import (
    OpcodeAPI,
)
from eth.tools._utils.deprecation import (
    deprecate_method,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.paris.opcodes import (
    PARIS_OPCODES,
)
from eth.vm.logic import (
    stack,
)
from eth.vm.opcode import (
    as_opcode,
)

from .logic import (
    Create2EIP3860,
    CreateEIP3860,
)

UPDATED_OPCODES: Dict[int, OpcodeAPI] = {
    opcode_values.CREATE: CreateEIP3860.configure(
        __name__="CreateEIP3860",
        mnemonic=mnemonics.CREATE,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.CREATE2: Create2EIP3860.configure(
        __name__="Create2EIP3860",
        mnemonic=mnemonics.CREATE2,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.SELFDESTRUCT: deprecate_method(
        PARIS_OPCODES[opcode_values.SELFDESTRUCT],
        message=(
            f"{mnemonics.SELFDESTRUCT} opcode present in computation. This opcode is "
            "deprecated and a breaking change to its functionality is likely to come "
            "in the future."
        ),
    ),
}
# deprecate_method() does not preserve attributes, so we need to manually copy them over.
_selfdestruct = PARIS_OPCODES[opcode_values.SELFDESTRUCT]
UPDATED_OPCODES[opcode_values.SELFDESTRUCT].mnemonic = _selfdestruct.mnemonic
UPDATED_OPCODES[opcode_values.SELFDESTRUCT].gas_cost = _selfdestruct.gas_cost


NEW_OPCODES: Dict[int, OpcodeAPI] = {
    opcode_values.PUSH0: as_opcode(
        logic_fn=stack.push0,
        mnemonic=mnemonics.PUSH0,
        gas_cost=constants.GAS_BASE,
    ),
}

SHANGHAI_OPCODES: Dict[int, OpcodeAPI] = merge(
    copy.deepcopy(PARIS_OPCODES),
    UPDATED_OPCODES,
    NEW_OPCODES,
)



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.forks.paris import ParisVM
from eth.vm.state import BaseState

from .blocks import ShanghaiBlock
from .headers import (
    configure_shanghai_header,
    create_shanghai_header_from_parent,
)
from .state import ShanghaiState


class ShanghaiVM(ParisVM):
    # fork name
    fork = "shanghai"

    # classes
    block_class: Type[BaseBlock] = ShanghaiBlock
    _state_class: Type[BaseState] = ShanghaiState

    # Methods
    create_header_from_parent = staticmethod(  # type: ignore
        create_shanghai_header_from_parent()
    )
    configure_header = configure_shanghai_header



from typing import (
    Dict,
)

from eth import (
    constants,
)
from eth.abc import (
    OpcodeAPI,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.logic import (
    arithmetic,
    block,
    call,
    comparison,
    context,
    duplication,
    flow,
    logging,
    memory,
    sha3,
    stack,
    storage,
    swap,
    system,
)
from eth.vm.opcode import (
    as_opcode,
)

FRONTIER_OPCODES: Dict[int, OpcodeAPI] = {
    #
    # Arithmetic
    #
    opcode_values.STOP: as_opcode(
        logic_fn=flow.stop,
        mnemonic=mnemonics.STOP,
        gas_cost=constants.GAS_ZERO,
    ),
    opcode_values.ADD: as_opcode(
        logic_fn=arithmetic.add,
        mnemonic=mnemonics.ADD,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.MUL: as_opcode(
        logic_fn=arithmetic.mul,
        mnemonic=mnemonics.MUL,
        gas_cost=constants.GAS_LOW,
    ),
    opcode_values.SUB: as_opcode(
        logic_fn=arithmetic.sub,
        mnemonic=mnemonics.SUB,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DIV: as_opcode(
        logic_fn=arithmetic.div,
        mnemonic=mnemonics.DIV,
        gas_cost=constants.GAS_LOW,
    ),
    opcode_values.SDIV: as_opcode(
        logic_fn=arithmetic.sdiv,
        mnemonic=mnemonics.SDIV,
        gas_cost=constants.GAS_LOW,
    ),
    opcode_values.MOD: as_opcode(
        logic_fn=arithmetic.mod,
        mnemonic=mnemonics.MOD,
        gas_cost=constants.GAS_LOW,
    ),
    opcode_values.SMOD: as_opcode(
        logic_fn=arithmetic.smod,
        mnemonic=mnemonics.SMOD,
        gas_cost=constants.GAS_LOW,
    ),
    opcode_values.ADDMOD: as_opcode(
        logic_fn=arithmetic.addmod,
        mnemonic=mnemonics.ADDMOD,
        gas_cost=constants.GAS_MID,
    ),
    opcode_values.MULMOD: as_opcode(
        logic_fn=arithmetic.mulmod,
        mnemonic=mnemonics.MULMOD,
        gas_cost=constants.GAS_MID,
    ),
    opcode_values.EXP: as_opcode(
        logic_fn=arithmetic.exp(gas_per_byte=constants.GAS_EXPBYTE),
        mnemonic=mnemonics.EXP,
        gas_cost=constants.GAS_EXP,
    ),
    opcode_values.SIGNEXTEND: as_opcode(
        logic_fn=arithmetic.signextend,
        mnemonic=mnemonics.SIGNEXTEND,
        gas_cost=constants.GAS_LOW,
    ),
    #
    # Comparisons
    #
    opcode_values.LT: as_opcode(
        logic_fn=comparison.lt,
        mnemonic=mnemonics.LT,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.GT: as_opcode(
        logic_fn=comparison.gt,
        mnemonic=mnemonics.GT,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SLT: as_opcode(
        logic_fn=comparison.slt,
        mnemonic=mnemonics.SLT,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SGT: as_opcode(
        logic_fn=comparison.sgt,
        mnemonic=mnemonics.SGT,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.EQ: as_opcode(
        logic_fn=comparison.eq,
        mnemonic=mnemonics.EQ,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.ISZERO: as_opcode(
        logic_fn=comparison.iszero,
        mnemonic=mnemonics.ISZERO,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.AND: as_opcode(
        logic_fn=comparison.and_op,
        mnemonic=mnemonics.AND,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.OR: as_opcode(
        logic_fn=comparison.or_op,
        mnemonic=mnemonics.OR,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.XOR: as_opcode(
        logic_fn=comparison.xor,
        mnemonic=mnemonics.XOR,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.NOT: as_opcode(
        logic_fn=comparison.not_op,
        mnemonic=mnemonics.NOT,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.BYTE: as_opcode(
        logic_fn=comparison.byte_op,
        mnemonic=mnemonics.BYTE,
        gas_cost=constants.GAS_VERYLOW,
    ),
    #
    # Sha3
    #
    opcode_values.SHA3: as_opcode(
        logic_fn=sha3.sha3,
        mnemonic=mnemonics.SHA3,
        gas_cost=constants.GAS_SHA3,
    ),
    #
    # Environment Information
    #
    opcode_values.ADDRESS: as_opcode(
        logic_fn=context.address,
        mnemonic=mnemonics.ADDRESS,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.BALANCE: as_opcode(
        logic_fn=context.balance,
        mnemonic=mnemonics.BALANCE,
        gas_cost=constants.GAS_BALANCE,
    ),
    opcode_values.ORIGIN: as_opcode(
        logic_fn=context.origin,
        mnemonic=mnemonics.ORIGIN,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.CALLER: as_opcode(
        logic_fn=context.caller,
        mnemonic=mnemonics.CALLER,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.CALLVALUE: as_opcode(
        logic_fn=context.callvalue,
        mnemonic=mnemonics.CALLVALUE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.CALLDATALOAD: as_opcode(
        logic_fn=context.calldataload,
        mnemonic=mnemonics.CALLDATALOAD,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.CALLDATASIZE: as_opcode(
        logic_fn=context.calldatasize,
        mnemonic=mnemonics.CALLDATASIZE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.CALLDATACOPY: as_opcode(
        logic_fn=context.calldatacopy,
        mnemonic=mnemonics.CALLDATACOPY,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.CODESIZE: as_opcode(
        logic_fn=context.codesize,
        mnemonic=mnemonics.CODESIZE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.CODECOPY: as_opcode(
        logic_fn=context.codecopy,
        mnemonic=mnemonics.CODECOPY,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.GASPRICE: as_opcode(
        logic_fn=context.gasprice,
        mnemonic=mnemonics.GASPRICE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.EXTCODESIZE: as_opcode(
        logic_fn=context.extcodesize,
        mnemonic=mnemonics.EXTCODESIZE,
        gas_cost=constants.GAS_EXTCODE,
    ),
    opcode_values.EXTCODECOPY: as_opcode(
        logic_fn=context.extcodecopy,
        mnemonic=mnemonics.EXTCODECOPY,
        gas_cost=constants.GAS_EXTCODE,
    ),
    #
    # Block Information
    #
    opcode_values.BLOCKHASH: as_opcode(
        logic_fn=block.blockhash,
        mnemonic=mnemonics.BLOCKHASH,
        gas_cost=constants.GAS_BLOCKHASH,
    ),
    opcode_values.COINBASE: as_opcode(
        logic_fn=block.coinbase,
        mnemonic=mnemonics.COINBASE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.TIMESTAMP: as_opcode(
        logic_fn=block.timestamp,
        mnemonic=mnemonics.TIMESTAMP,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.NUMBER: as_opcode(
        logic_fn=block.number,
        mnemonic=mnemonics.NUMBER,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.DIFFICULTY: as_opcode(
        logic_fn=block.difficulty,
        mnemonic=mnemonics.DIFFICULTY,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.GASLIMIT: as_opcode(
        logic_fn=block.gaslimit,
        mnemonic=mnemonics.GASLIMIT,
        gas_cost=constants.GAS_BASE,
    ),
    #
    # Stack, Memory, Storage and Flow Operations
    #
    opcode_values.POP: as_opcode(
        logic_fn=stack.pop,
        mnemonic=mnemonics.POP,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.MLOAD: as_opcode(
        logic_fn=memory.mload,
        mnemonic=mnemonics.MLOAD,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.MSTORE: as_opcode(
        logic_fn=memory.mstore,
        mnemonic=mnemonics.MSTORE,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.MSTORE8: as_opcode(
        logic_fn=memory.mstore8,
        mnemonic=mnemonics.MSTORE8,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SLOAD: as_opcode(
        logic_fn=storage.sload,
        mnemonic=mnemonics.SLOAD,
        gas_cost=constants.GAS_SLOAD,
    ),
    opcode_values.SSTORE: as_opcode(
        logic_fn=storage.sstore,
        mnemonic=mnemonics.SSTORE,
        gas_cost=constants.GAS_NULL,
    ),
    opcode_values.JUMP: as_opcode(
        logic_fn=flow.jump,
        mnemonic=mnemonics.JUMP,
        gas_cost=constants.GAS_MID,
    ),
    opcode_values.JUMPI: as_opcode(
        logic_fn=flow.jumpi,
        mnemonic=mnemonics.JUMPI,
        gas_cost=constants.GAS_HIGH,
    ),
    opcode_values.PC: as_opcode(
        logic_fn=flow.program_counter,
        mnemonic=mnemonics.PC,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.MSIZE: as_opcode(
        logic_fn=memory.msize,
        mnemonic=mnemonics.MSIZE,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.GAS: as_opcode(
        logic_fn=flow.gas,
        mnemonic=mnemonics.GAS,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.JUMPDEST: as_opcode(
        logic_fn=flow.jumpdest,
        mnemonic=mnemonics.JUMPDEST,
        gas_cost=constants.GAS_JUMPDEST,
    ),
    #
    # Push Operations
    #
    opcode_values.PUSH1: as_opcode(
        logic_fn=stack.push1,
        mnemonic=mnemonics.PUSH1,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH2: as_opcode(
        logic_fn=stack.push2,
        mnemonic=mnemonics.PUSH2,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH3: as_opcode(
        logic_fn=stack.push3,
        mnemonic=mnemonics.PUSH3,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH4: as_opcode(
        logic_fn=stack.push4,
        mnemonic=mnemonics.PUSH4,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH5: as_opcode(
        logic_fn=stack.push5,
        mnemonic=mnemonics.PUSH5,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH6: as_opcode(
        logic_fn=stack.push6,
        mnemonic=mnemonics.PUSH6,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH7: as_opcode(
        logic_fn=stack.push7,
        mnemonic=mnemonics.PUSH7,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH8: as_opcode(
        logic_fn=stack.push8,
        mnemonic=mnemonics.PUSH8,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH9: as_opcode(
        logic_fn=stack.push9,
        mnemonic=mnemonics.PUSH9,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH10: as_opcode(
        logic_fn=stack.push10,
        mnemonic=mnemonics.PUSH10,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH11: as_opcode(
        logic_fn=stack.push11,
        mnemonic=mnemonics.PUSH11,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH12: as_opcode(
        logic_fn=stack.push12,
        mnemonic=mnemonics.PUSH12,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH13: as_opcode(
        logic_fn=stack.push13,
        mnemonic=mnemonics.PUSH13,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH14: as_opcode(
        logic_fn=stack.push14,
        mnemonic=mnemonics.PUSH14,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH15: as_opcode(
        logic_fn=stack.push15,
        mnemonic=mnemonics.PUSH15,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH16: as_opcode(
        logic_fn=stack.push16,
        mnemonic=mnemonics.PUSH16,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH17: as_opcode(
        logic_fn=stack.push17,
        mnemonic=mnemonics.PUSH17,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH18: as_opcode(
        logic_fn=stack.push18,
        mnemonic=mnemonics.PUSH18,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH19: as_opcode(
        logic_fn=stack.push19,
        mnemonic=mnemonics.PUSH19,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH20: as_opcode(
        logic_fn=stack.push20,
        mnemonic=mnemonics.PUSH20,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH21: as_opcode(
        logic_fn=stack.push21,
        mnemonic=mnemonics.PUSH21,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH22: as_opcode(
        logic_fn=stack.push22,
        mnemonic=mnemonics.PUSH22,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH23: as_opcode(
        logic_fn=stack.push23,
        mnemonic=mnemonics.PUSH23,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH24: as_opcode(
        logic_fn=stack.push24,
        mnemonic=mnemonics.PUSH24,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH25: as_opcode(
        logic_fn=stack.push25,
        mnemonic=mnemonics.PUSH25,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH26: as_opcode(
        logic_fn=stack.push26,
        mnemonic=mnemonics.PUSH26,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH27: as_opcode(
        logic_fn=stack.push27,
        mnemonic=mnemonics.PUSH27,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH28: as_opcode(
        logic_fn=stack.push28,
        mnemonic=mnemonics.PUSH28,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH29: as_opcode(
        logic_fn=stack.push29,
        mnemonic=mnemonics.PUSH29,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH30: as_opcode(
        logic_fn=stack.push30,
        mnemonic=mnemonics.PUSH30,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH31: as_opcode(
        logic_fn=stack.push31,
        mnemonic=mnemonics.PUSH31,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.PUSH32: as_opcode(
        logic_fn=stack.push32,
        mnemonic=mnemonics.PUSH32,
        gas_cost=constants.GAS_VERYLOW,
    ),
    #
    # Duplicate Operations
    #
    opcode_values.DUP1: as_opcode(
        logic_fn=duplication.dup1,
        mnemonic=mnemonics.DUP1,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP2: as_opcode(
        logic_fn=duplication.dup2,
        mnemonic=mnemonics.DUP2,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP3: as_opcode(
        logic_fn=duplication.dup3,
        mnemonic=mnemonics.DUP3,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP4: as_opcode(
        logic_fn=duplication.dup4,
        mnemonic=mnemonics.DUP4,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP5: as_opcode(
        logic_fn=duplication.dup5,
        mnemonic=mnemonics.DUP5,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP6: as_opcode(
        logic_fn=duplication.dup6,
        mnemonic=mnemonics.DUP6,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP7: as_opcode(
        logic_fn=duplication.dup7,
        mnemonic=mnemonics.DUP7,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP8: as_opcode(
        logic_fn=duplication.dup8,
        mnemonic=mnemonics.DUP8,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP9: as_opcode(
        logic_fn=duplication.dup9,
        mnemonic=mnemonics.DUP9,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP10: as_opcode(
        logic_fn=duplication.dup10,
        mnemonic=mnemonics.DUP10,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP11: as_opcode(
        logic_fn=duplication.dup11,
        mnemonic=mnemonics.DUP11,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP12: as_opcode(
        logic_fn=duplication.dup12,
        mnemonic=mnemonics.DUP12,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP13: as_opcode(
        logic_fn=duplication.dup13,
        mnemonic=mnemonics.DUP13,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP14: as_opcode(
        logic_fn=duplication.dup14,
        mnemonic=mnemonics.DUP14,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP15: as_opcode(
        logic_fn=duplication.dup15,
        mnemonic=mnemonics.DUP15,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.DUP16: as_opcode(
        logic_fn=duplication.dup16,
        mnemonic=mnemonics.DUP16,
        gas_cost=constants.GAS_VERYLOW,
    ),
    #
    # Exchange Operations
    #
    opcode_values.SWAP1: as_opcode(
        logic_fn=swap.swap1,
        mnemonic=mnemonics.SWAP1,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP2: as_opcode(
        logic_fn=swap.swap2,
        mnemonic=mnemonics.SWAP2,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP3: as_opcode(
        logic_fn=swap.swap3,
        mnemonic=mnemonics.SWAP3,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP4: as_opcode(
        logic_fn=swap.swap4,
        mnemonic=mnemonics.SWAP4,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP5: as_opcode(
        logic_fn=swap.swap5,
        mnemonic=mnemonics.SWAP5,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP6: as_opcode(
        logic_fn=swap.swap6,
        mnemonic=mnemonics.SWAP6,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP7: as_opcode(
        logic_fn=swap.swap7,
        mnemonic=mnemonics.SWAP7,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP8: as_opcode(
        logic_fn=swap.swap8,
        mnemonic=mnemonics.SWAP8,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP9: as_opcode(
        logic_fn=swap.swap9,
        mnemonic=mnemonics.SWAP9,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP10: as_opcode(
        logic_fn=swap.swap10,
        mnemonic=mnemonics.SWAP10,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP11: as_opcode(
        logic_fn=swap.swap11,
        mnemonic=mnemonics.SWAP11,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP12: as_opcode(
        logic_fn=swap.swap12,
        mnemonic=mnemonics.SWAP12,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP13: as_opcode(
        logic_fn=swap.swap13,
        mnemonic=mnemonics.SWAP13,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP14: as_opcode(
        logic_fn=swap.swap14,
        mnemonic=mnemonics.SWAP14,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP15: as_opcode(
        logic_fn=swap.swap15,
        mnemonic=mnemonics.SWAP15,
        gas_cost=constants.GAS_VERYLOW,
    ),
    opcode_values.SWAP16: as_opcode(
        logic_fn=swap.swap16,
        mnemonic=mnemonics.SWAP16,
        gas_cost=constants.GAS_VERYLOW,
    ),
    #
    # Logging
    #
    opcode_values.LOG0: as_opcode(
        logic_fn=logging.log0,
        mnemonic=mnemonics.LOG0,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG1: as_opcode(
        logic_fn=logging.log1,
        mnemonic=mnemonics.LOG1,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG2: as_opcode(
        logic_fn=logging.log2,
        mnemonic=mnemonics.LOG2,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG3: as_opcode(
        logic_fn=logging.log3,
        mnemonic=mnemonics.LOG3,
        gas_cost=constants.GAS_LOG,
    ),
    opcode_values.LOG4: as_opcode(
        logic_fn=logging.log4,
        mnemonic=mnemonics.LOG4,
        gas_cost=constants.GAS_LOG,
    ),
    #
    # System
    #
    opcode_values.CREATE: system.Create.configure(
        __name__="opcode:CREATE",
        mnemonic=mnemonics.CREATE,
        gas_cost=constants.GAS_CREATE,
    )(),
    opcode_values.CALL: call.Call.configure(
        __name__="opcode:CALL",
        mnemonic=mnemonics.CALL,
        gas_cost=constants.GAS_CALL,
    )(),
    opcode_values.CALLCODE: call.CallCode.configure(
        __name__="opcode:CALLCODE",
        mnemonic=mnemonics.CALLCODE,
        gas_cost=constants.GAS_CALL,
    )(),
    opcode_values.RETURN: as_opcode(
        logic_fn=system.return_op,
        mnemonic=mnemonics.RETURN,
        gas_cost=constants.GAS_ZERO,
    ),
    opcode_values.SELFDESTRUCT: as_opcode(
        logic_fn=system.selfdestruct,
        mnemonic=mnemonics.SELFDESTRUCT,
        gas_cost=constants.GAS_SELFDESTRUCT,
    ),
}



from typing import (
    Optional,
)

from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    encode_hex,
)
from trie.exceptions import (
    MissingTrieNode,
)
from trie.typing import (
    Nibbles,
)

from eth.exceptions import (
    PyEVMError,
)


class EVMMissingData(PyEVMError):
    pass


class MissingAccountTrieNode(EVMMissingData, MissingTrieNode):
    """
    Raised when a main state trie node is missing from the DB, to get an account RLP
    """

    @property
    def state_root_hash(self) -> Hash32:
        return self.root_hash

    @property
    def address_hash(self) -> Hash32:
        return self.requested_key

    def __repr__(self) -> str:
        return f"MissingAccountTrieNode: {self}"

    def __str__(self) -> str:
        superclass_str = EVMMissingData.__str__(self)
        return (
            "State trie database is missing node for hash "
            f"{encode_hex(self.missing_node_hash)}, which is needed to look up account "
            f"with address hash {encode_hex(self.address_hash)} "
            f"at root hash {encode_hex(self.state_root_hash)} "
            f"-- {superclass_str}"
        )


class MissingStorageTrieNode(EVMMissingData, MissingTrieNode):
    """
    Raised when a storage trie node is missing from the DB
    """

    def __init__(
        self,
        missing_node_hash: Hash32,
        storage_root_hash: Hash32,
        requested_key: Hash32,
        prefix: Optional[Nibbles],
        account_address: Address,
        *args: bytes,
    ) -> None:
        if not isinstance(account_address, bytes):
            raise TypeError(f"Account address must be bytes, was: {account_address!r}")

        super().__init__(
            missing_node_hash,
            storage_root_hash,
            requested_key,
            prefix,
            account_address,
            *args,
        )

    @property
    def storage_root_hash(self) -> Hash32:
        return self.root_hash

    @property
    def account_address(self) -> Address:
        return self.args[4]

    def __repr__(self) -> str:
        return f"MissingStorageTrieNode: {self}"

    def __str__(self) -> str:
        superclass_str = EVMMissingData.__str__(self)
        return (
            "Storage trie database is missing hash "
            f"{encode_hex(self.missing_node_hash)} "
            f"needed to look up key {encode_hex(self.requested_key)} "
            f"at root hash {encode_hex(self.root_hash)} "
            f"in account address {encode_hex(self.account_address)} "
            f"-- {superclass_str}"
        )


class MissingBytecode(EVMMissingData):
    """
    Raised when the bytecode is missing from the database for a known bytecode hash.
    """

    def __init__(self, missing_code_hash: Hash32) -> None:
        if not isinstance(missing_code_hash, bytes):
            raise TypeError(
                f"Missing code hash must be bytes, was: {missing_code_hash!r}"
            )

        super().__init__(missing_code_hash)

    @property
    def missing_code_hash(self) -> Hash32:
        return self.args[0]

    def __repr__(self) -> str:
        return f"MissingBytecode: {self}"

    def __str__(self) -> str:
        superclass_str = EVMMissingData.__str__(self)
        return (
            f"Database is missing bytecode for code hash "
            f"{encode_hex(self.missing_code_hash)} "
            f"-- {superclass_str}"
        )



import contextlib
from typing import (
    Iterator,
    Sequence,
    Tuple,
    Type,
)

from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)
from eth_utils import (
    ExtendedDebugLogger,
    get_extended_debug_logger,
)
from eth_utils.toolz import (
    nth,
)

from eth._utils.datatypes import (
    Configurable,
)
from eth.abc import (
    AccountDatabaseAPI,
    AtomicDatabaseAPI,
    ComputationAPI,
    ExecutionContextAPI,
    MessageAPI,
    MetaWitnessAPI,
    SignedTransactionAPI,
    StateAPI,
    TransactionContextAPI,
    TransactionExecutorAPI,
    WithdrawalAPI,
)
from eth.constants import (
    MAX_PREV_HEADER_DEPTH,
)
from eth.typing import (
    JournalDBCheckpoint,
)
from web3 import Web3
from web3.types import (
    BlockIdentifier,
)

class BaseState(Configurable, StateAPI):
    #
    # Set from __init__
    #
    __slots__ = ["_db", "execution_context", "_account_db"]

    computation_class: Type[ComputationAPI] = None
    transaction_context_class: Type[TransactionContextAPI] = None
    account_db_class: Type[AccountDatabaseAPI] = None
    transaction_executor_class: Type[TransactionExecutorAPI] = None

    def __init__(
        self,
        web3: Web3,
        block_identifier: BlockIdentifier,
        execution_context: ExecutionContextAPI,
    ) -> None:
        self.execution_context = execution_context
        self._account_db = self.get_account_db_class()(web3, block_identifier)

    #
    # Logging
    #
    @property
    def logger(self) -> ExtendedDebugLogger:
        return get_extended_debug_logger(f"eth.vm.state.{self.__class__.__name__}")

    #
    # Block Object Properties (in opcodes)
    #

    @property
    def coinbase(self) -> Address:
        return self.execution_context.coinbase

    @property
    def timestamp(self) -> int:
        return self.execution_context.timestamp

    @property
    def block_number(self) -> BlockNumber:
        return self.execution_context.block_number

    @property
    def difficulty(self) -> int:
        return self.execution_context.difficulty

    @property
    def mix_hash(self) -> Hash32:
        return self.execution_context.mix_hash

    @property
    def gas_limit(self) -> int:
        return self.execution_context.gas_limit

    @property
    def base_fee(self) -> int:
        raise NotImplementedError(
            "Basefee opcode is not implemented prior to London hard fork"
        )

    def get_tip(self, transaction: SignedTransactionAPI) -> int:
        return transaction.gas_price

    def get_gas_price(self, transaction: SignedTransactionAPI) -> int:
        return transaction.gas_price

    #
    # Access to account db
    #
    @classmethod
    def get_account_db_class(cls) -> Type[AccountDatabaseAPI]:
        if cls.account_db_class is None:
            raise AttributeError(f"No account_db_class set for {cls.__name__}")
        return cls.account_db_class

    @property
    def state_root(self) -> Hash32:
        return self._account_db.state_root

    def make_state_root(self) -> Hash32:
        return self._account_db.make_state_root()

    def get_storage(
        self, address: Address, slot: int, from_journal: bool = True
    ) -> int:
        return self._account_db.get_storage(address, slot, from_journal)

    def set_storage(self, address: Address, slot: int, value: int) -> None:
        return self._account_db.set_storage(address, slot, value)

    def delete_storage(self, address: Address) -> None:
        self._account_db.delete_storage(address)

    def delete_account(self, address: Address) -> None:
        self._account_db.delete_account(address)

    def get_balance(self, address: Address) -> int:
        return self._account_db.get_balance(address)

    def set_balance(self, address: Address, balance: int) -> None:
        self._account_db.set_balance(address, balance)

    def delta_balance(self, address: Address, delta: int) -> None:
        self.set_balance(address, self.get_balance(address) + delta)

    def get_nonce(self, address: Address) -> int:
        return self._account_db.get_nonce(address)

    def set_nonce(self, address: Address, nonce: int) -> None:
        self._account_db.set_nonce(address, nonce)

    def increment_nonce(self, address: Address) -> None:
        self._account_db.increment_nonce(address)

    def get_code(self, address: Address) -> bytes:
        return self._account_db.get_code(address)

    def set_code(self, address: Address, code: bytes) -> None:
        self._account_db.set_code(address, code)

    def get_code_hash(self, address: Address) -> Hash32:
        return self._account_db.get_code_hash(address)

    def delete_code(self, address: Address) -> None:
        self._account_db.delete_code(address)

    def has_code_or_nonce(self, address: Address) -> bool:
        return self._account_db.account_has_code_or_nonce(address)

    def account_exists(self, address: Address) -> bool:
        return self._account_db.account_exists(address)

    def touch_account(self, address: Address) -> None:
        self._account_db.touch_account(address)

    def account_is_empty(self, address: Address) -> bool:
        return self._account_db.account_is_empty(address)

    def is_storage_warm(self, address: Address, slot: int) -> bool:
        return self._account_db.is_storage_warm(address, slot)

    def mark_storage_warm(self, address: Address, slot: int) -> None:
        return self._account_db.mark_storage_warm(address, slot)

    def is_address_warm(self, address: Address) -> bool:
        """
        Was the account accessed during this transaction?

        See EIP-2929
        """
        return (
            self._account_db.is_address_warm(address)
            or address in self.computation_class.get_precompiles()
        )

    def mark_address_warm(self, address: Address) -> None:
        self._account_db.mark_address_warm(address)

    #
    # Access self._chaindb
    #
    def snapshot(self) -> Tuple[JournalDBCheckpoint,]:
        return (self._account_db.record(),)

    def revert(self, snapshot: Tuple[JournalDBCheckpoint,]) -> None:
        (account_snapshot,) = snapshot

        # now roll the underlying database back
        self._account_db.discard(account_snapshot)

    def commit(self, snapshot: Tuple[JournalDBCheckpoint]) -> None:
        (account_snapshot,) = snapshot
        self._account_db.commit(account_snapshot)

    def lock_changes(self) -> None:
        self._account_db.lock_changes()

    def persist(self) -> MetaWitnessAPI:
        return self._account_db.persist()

    #
    # Access self.prev_hashes (Read-only)
    #
    def get_ancestor_hash(self, block_number: int) -> Hash32:
        ancestor_depth = self.block_number - block_number - 1
        is_ancestor_depth_out_of_range = (
            ancestor_depth >= MAX_PREV_HEADER_DEPTH
            or ancestor_depth < 0
            or block_number < 0
        )
        if is_ancestor_depth_out_of_range:
            return Hash32(b"")

        try:
            return nth(ancestor_depth, self.execution_context.prev_hashes)
        except StopIteration:
            # Ancestor with specified depth not present
            return Hash32(b"")

    #
    # Computation
    #
    def get_computation(
        self, message: MessageAPI, transaction_context: TransactionContextAPI
    ) -> ComputationAPI:
        if self.computation_class is None:
            raise AttributeError("No `computation_class` has been set for this State")
        else:
            computation = self.computation_class(self, message, transaction_context)
        return computation

    #
    # Transaction context
    #
    @classmethod
    def get_transaction_context_class(cls) -> Type[TransactionContextAPI]:
        if cls.transaction_context_class is None:
            raise AttributeError(
                "No `transaction_context_class` has been set for this State"
            )
        return cls.transaction_context_class

    #
    # Execution
    #
    def get_transaction_executor(self) -> TransactionExecutorAPI:
        return self.transaction_executor_class(self)

    def costless_execute_transaction(
        self, transaction: SignedTransactionAPI
    ) -> ComputationAPI:
        with self.override_transaction_context(gas_price=transaction.gas_price):
            free_transaction = transaction.copy(gas_price=0)
            return self.apply_transaction(free_transaction)

    @contextlib.contextmanager
    def override_transaction_context(self, gas_price: int) -> Iterator[None]:
        original_context = self.get_transaction_context

        def get_custom_transaction_context(
            transaction: SignedTransactionAPI,
        ) -> TransactionContextAPI:
            custom_transaction = transaction.copy(gas_price=gas_price)
            return original_context(custom_transaction)

        # mypy doesn't like assigning to an existing method
        self.get_transaction_context = get_custom_transaction_context  # type: ignore
        try:
            yield
        finally:
            self.get_transaction_context = original_context  # type: ignore # Remove ignore if https://github.com/python/mypy/issues/708 is fixed. # noqa: E501

    def get_transaction_context(
        self, transaction: SignedTransactionAPI
    ) -> TransactionContextAPI:
        return self.get_transaction_context_class()(
            gas_price=transaction.gas_price,
            origin=transaction.sender,
        )

    #
    # Withdrawals
    #

    def apply_withdrawal(self, withdrawal: WithdrawalAPI) -> None:
        # withdrawals not implemented until the Shanghai hard fork
        pass

    def apply_all_withdrawals(self, withdrawals: Sequence[WithdrawalAPI]) -> None:
        # withdrawals not implemented until the Shanghai hard fork
        pass


class BaseTransactionExecutor(TransactionExecutorAPI):
    def __init__(self, vm_state: StateAPI) -> None:
        self.vm_state = vm_state

    def __call__(self, transaction: SignedTransactionAPI) -> ComputationAPI:
        self.validate_transaction(transaction)
        message = self.build_evm_message(transaction)
        computation = self.build_computation(message, transaction)
        finalized_computation = self.finalize_computation(transaction, computation)
        return finalized_computation



from eth_hash.auto import (
    keccak,
)
from eth_utils import (
    encode_hex,
)

from eth import (
    constants,
)
from eth.abc import (
    ComputationAPI,
    MessageAPI,
    StateAPI,
    TransactionContextAPI,
)
from eth.exceptions import (
    OutOfGas,
    VMError,
)
from eth.vm.forks.homestead.computation import (
    HomesteadComputation,
)

from ..spurious_dragon.constants import (
    EIP170_CODE_SIZE_LIMIT,
)
from .opcodes import (
    SPURIOUS_DRAGON_OPCODES,
)


class SpuriousDragonComputation(HomesteadComputation):
    """
    A class for all execution *message* computations in the ``SpuriousDragon`` fork.
    Inherits from
    :class:`~eth.vm.forks.homestead.computation.HomesteadComputation`
    """

    # Override
    opcodes = SPURIOUS_DRAGON_OPCODES

    @classmethod
    def apply_create_message(
        cls,
        state: StateAPI,
        message: MessageAPI,
        transaction_context: TransactionContextAPI,
    ) -> ComputationAPI:
        snapshot = state.snapshot()

        # EIP161 nonce incrementation
        state.increment_nonce(message.storage_address)

        cls.validate_create_message(message)

        computation = cls.apply_message(state, message, transaction_context)

        if computation.is_error:
            state.revert(snapshot)
            return computation
        else:
            contract_code = computation.output

            if contract_code:
                try:
                    cls.validate_contract_code(contract_code)

                    contract_code_gas_cost = (
                        len(contract_code) * constants.GAS_CODEDEPOSIT
                    )
                    computation.consume_gas(
                        contract_code_gas_cost,
                        reason="Write contract code for CREATE",
                    )
                except VMError as err:
                    # Different from Frontier: reverts state on gas failure while
                    # writing contract code.
                    computation.error = err
                    state.revert(snapshot)
                    cls.logger.debug2(f"VMError setting contract code: {err}")
                else:
                    if cls.logger:
                        cls.logger.debug2(
                            "SETTING CODE: %s -> length: %s | hash: %s",
                            encode_hex(message.storage_address),
                            len(contract_code),
                            encode_hex(keccak(contract_code)),
                        )

                    state.set_code(message.storage_address, contract_code)
                    state.commit(snapshot)
            else:
                state.commit(snapshot)
            return computation

    @classmethod
    def validate_create_message(cls, message: MessageAPI) -> None:
        # this method does not become relevant until the Shanghai hard fork
        """
        Class method for validating a create message.
        """
        pass

    @classmethod
    def validate_contract_code(cls, contract_code: bytes) -> None:
        if len(contract_code) > EIP170_CODE_SIZE_LIMIT:
            raise OutOfGas(
                f"Contract code size exceeds EIP170 limit of {EIP170_CODE_SIZE_LIMIT}."
                f"  Got code of size: {len(contract_code)}"
            )



from typing import (
    Tuple,
    Type,
    cast,
)

from eth_typing import (
    BlockNumber,
    Hash32,
)

from eth._utils.datatypes import (
    Configurable,
)
from eth.abc import (
    AtomicDatabaseAPI,
    BlockHeaderAPI,
    HeaderChainAPI,
    HeaderDatabaseAPI,
)
from eth.db.backends.base import (
    BaseAtomicDB,
)
from eth.db.header import (
    HeaderDB,
)


class HeaderChain(Configurable, HeaderChainAPI):
    _base_db: AtomicDatabaseAPI = None
    _headerdb: HeaderDatabaseAPI = None

    _headerdb_class: Type[HeaderDatabaseAPI] = HeaderDB

    def __init__(
        self, base_db: AtomicDatabaseAPI, header: BlockHeaderAPI = None
    ) -> None:
        self.base_db = base_db
        self.headerdb = self.get_headerdb_class()(base_db)

        if header is None:
            self.header = self.get_canonical_head()
        else:
            self.header = header

    #
    # Chain Initialization API
    #
    @classmethod
    def from_genesis_header(
        cls, base_db: AtomicDatabaseAPI, genesis_header: BlockHeaderAPI
    ) -> HeaderChainAPI:
        headerdb = cls.get_headerdb_class()(cast(BaseAtomicDB, base_db))
        headerdb.persist_header(genesis_header)
        return cls(base_db, genesis_header)

    #
    # Helpers
    #
    @classmethod
    def get_headerdb_class(cls) -> Type[HeaderDatabaseAPI]:
        if cls._headerdb_class is None:
            raise AttributeError("HeaderChain classes must set a `headerdb_class`")
        return cls._headerdb_class

    #
    # Canonical Chain API
    #
    def get_canonical_block_hash(self, block_number: BlockNumber) -> Hash32:
        return self.headerdb.get_canonical_block_hash(block_number)

    def get_canonical_block_header_by_number(
        self, block_number: BlockNumber
    ) -> BlockHeaderAPI:
        return self.headerdb.get_canonical_block_header_by_number(block_number)

    def get_canonical_head(self) -> BlockHeaderAPI:
        return self.headerdb.get_canonical_head()

    #
    # Header API
    #
    def get_block_header_by_hash(self, block_hash: Hash32) -> BlockHeaderAPI:
        return self.headerdb.get_block_header_by_hash(block_hash)

    def header_exists(self, block_hash: Hash32) -> bool:
        return self.headerdb.header_exists(block_hash)

    def import_header(
        self, header: BlockHeaderAPI
    ) -> Tuple[Tuple[BlockHeaderAPI, ...], Tuple[BlockHeaderAPI, ...]]:
        new_canonical_headers = self.headerdb.persist_header(header)
        self.header = self.get_canonical_head()
        return new_canonical_headers



from .loading import (  # noqa: F401
    find_fixtures,
    filter_fixtures,
    load_fixture,
)
from .generation import (  # noqa: F401
    generate_fixture_tests,
)
from .helpers import (  # noqa: F401
    new_chain_from_fixture,
    genesis_fields_from_fixture,
    genesis_params_from_fixture,
    apply_fixture_block_to_chain,
    setup_state,
    should_run_slow_tests,
    verify_state,
)
from eth.tools._utils.normalization import (  # noqa: F401
    normalize_block,
    normalize_blockchain_fixtures,
    normalize_statetest_fixture,
    normalize_transactiontest_fixture,
    normalize_vmtest_fixture,
)



from typing import (
    Dict,
    Iterator,
)

from .base import (
    BaseDB,
)


class MemoryDB(BaseDB):
    kv_store: Dict[bytes, bytes] = None

    def __init__(self, kv_store: Dict[bytes, bytes] = None) -> None:
        if kv_store is None:
            self.kv_store = {}
        else:
            self.kv_store = kv_store

    def __getitem__(self, key: bytes) -> bytes:
        return self.kv_store[key]

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self.kv_store[key] = value

    def _exists(self, key: bytes) -> bool:
        return key in self.kv_store

    def __delitem__(self, key: bytes) -> None:
        del self.kv_store[key]

    def __iter__(self) -> Iterator[bytes]:
        return iter(self.kv_store)

    def __len__(self) -> int:
        return len(self.kv_store)

    def __repr__(self) -> str:
        return f"MemoryDB({self.kv_store!r})"



from abc import (
    ABC,
)
from typing import (
    List,
    Sequence,
    Tuple,
    Type,
    cast,
)

from eth_bloom import (
    BloomFilter,
)
from eth_typing import (
    Address,
    BlockNumber,
    Hash32,
)
from eth_utils import (
    encode_hex,
    keccak,
)
import rlp
from rlp.sedes import (
    Binary,
    CountableList,
    big_endian_int,
    binary,
)
from trie.exceptions import (
    MissingTrieNode,
)

from eth._utils.headers import (
    new_timestamp_from_parent,
)
from eth.abc import (
    BlockHeaderAPI,
    BlockHeaderSedesAPI,
    ChainDatabaseAPI,
    MiningHeaderAPI,
    ReceiptAPI,
    ReceiptBuilderAPI,
    SignedTransactionAPI,
    TransactionBuilderAPI,
    WithdrawalAPI,
)
from eth.constants import (
    BLANK_ROOT_HASH,
    EMPTY_UNCLE_HASH,
    GENESIS_NONCE,
    GENESIS_PARENT_HASH,
    ZERO_ADDRESS,
    ZERO_HASH32,
)
from eth.exceptions import (
    BlockNotFound,
    HeaderNotFound,
)
from eth.rlp.blocks import (
    BaseBlock,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.rlp.sedes import (
    address,
    hash32,
    trie_root,
    uint256,
)

from ..london.blocks import (
    LondonBlockHeader,
)
from ..london.receipts import (
    LondonReceiptBuilder,
)
from .transactions import (
    ShanghaiTransactionBuilder,
)
from .withdrawals import (
    Withdrawal,
)

UNMINED_SHANGHAI_HEADER_FIELDS = [
    ("parent_hash", hash32),
    ("uncles_hash", hash32),
    ("coinbase", address),
    ("state_root", trie_root),
    ("transaction_root", trie_root),
    ("receipt_root", trie_root),
    ("bloom", uint256),
    ("difficulty", big_endian_int),
    ("block_number", big_endian_int),
    ("gas_limit", big_endian_int),
    ("gas_used", big_endian_int),
    ("timestamp", big_endian_int),
    ("extra_data", binary),
    ("base_fee_per_gas", big_endian_int),
    ("withdrawals_root", trie_root),
]


class ShanghaiMiningHeader(rlp.Serializable, MiningHeaderAPI, ABC):
    fields = UNMINED_SHANGHAI_HEADER_FIELDS


class ShanghaiBlockHeader(rlp.Serializable, BlockHeaderAPI, ABC):
    # `mix_hash` and `nonce` were fields before `base_fee_per_gas` and
    # `withdrawals_root` and, thus, appear in the block header before them.
    fields = (
        UNMINED_SHANGHAI_HEADER_FIELDS[:13]
        + [
            ("mix_hash", binary),
            ("nonce", Binary(8, allow_empty=True)),
        ]
        + UNMINED_SHANGHAI_HEADER_FIELDS[13:]
    )

    def __init__(
        self,
        difficulty: int,
        block_number: BlockNumber,
        gas_limit: int,
        timestamp: int = None,
        coinbase: Address = ZERO_ADDRESS,
        parent_hash: Hash32 = ZERO_HASH32,
        uncles_hash: Hash32 = EMPTY_UNCLE_HASH,
        state_root: Hash32 = BLANK_ROOT_HASH,
        transaction_root: Hash32 = BLANK_ROOT_HASH,
        receipt_root: Hash32 = BLANK_ROOT_HASH,
        bloom: int = 0,
        gas_used: int = 0,
        extra_data: bytes = b"",
        mix_hash: Hash32 = ZERO_HASH32,
        nonce: bytes = GENESIS_NONCE,
        base_fee_per_gas: int = 0,
        withdrawals_root: Hash32 = BLANK_ROOT_HASH,
    ) -> None:
        if timestamp is None:
            if parent_hash == ZERO_HASH32:
                timestamp = new_timestamp_from_parent(None)
            else:
                # without access to the parent header, we cannot select a new
                # timestamp correctly
                raise ValueError(
                    "Must set timestamp explicitly if this is not a genesis header"
                )

        super().__init__(
            parent_hash=parent_hash,
            uncles_hash=uncles_hash,
            coinbase=coinbase,
            state_root=state_root,
            transaction_root=transaction_root,
            receipt_root=receipt_root,
            bloom=bloom,
            difficulty=difficulty,
            block_number=block_number,
            gas_limit=gas_limit,
            gas_used=gas_used,
            timestamp=timestamp,
            extra_data=extra_data,
            mix_hash=mix_hash,
            nonce=nonce,
            base_fee_per_gas=base_fee_per_gas,
            withdrawals_root=withdrawals_root,
        )

    def __str__(self) -> str:
        return (
            f"<ShanghaiBlockHeader "
            f"#{self.block_number} {encode_hex(self.hash)[2:10]}>"
        )

    _hash = None

    @property
    def hash(self) -> Hash32:
        if self._hash is None:
            self._hash = keccak(rlp.encode(self))
        return cast(Hash32, self._hash)

    @property
    def mining_hash(self) -> Hash32:
        non_pow_fields = self[:-3] + self[-1:]
        result = keccak(rlp.encode(non_pow_fields, ShanghaiMiningHeader))
        return cast(Hash32, result)

    @property
    def hex_hash(self) -> str:
        return encode_hex(self.hash)

    @property
    def is_genesis(self) -> bool:
        return self.parent_hash == GENESIS_PARENT_HASH and self.block_number == 0


class ShanghaiBackwardsHeader(BlockHeaderSedesAPI):
    """
    An rlp sedes class for block headers.
    It can serialize and deserialize Shanghai, London, and pre-London headers.
    """

    @classmethod
    def serialize(cls, obj: BlockHeaderAPI) -> List[bytes]:
        return obj.serialize(obj)

    @classmethod
    def deserialize(cls, encoded: List[bytes]) -> BlockHeaderAPI:
        num_fields = len(encoded)
        if num_fields == 17:
            return ShanghaiBlockHeader.deserialize(encoded)
        if num_fields == 16:
            return LondonBlockHeader.deserialize(encoded)
        elif num_fields == 15:
            return BlockHeader.deserialize(encoded)
        else:
            raise ValueError(
                "Unexpected number of fields in block header."
                f"Got {num_fields} in {encoded!r}"
            )


class ShanghaiBlock(BaseBlock):
    # re-defined from `BaseBlock`, as `FrontierBlock` was, to include withdrawals

    transaction_builder: Type[TransactionBuilderAPI] = ShanghaiTransactionBuilder
    # London was the last fork where the receipt builder was updated
    receipt_builder: Type[ReceiptBuilderAPI] = LondonReceiptBuilder
    fields = [
        ("header", ShanghaiBlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(ShanghaiBackwardsHeader, max_length=0)),
        ("withdrawals", CountableList(Withdrawal)),
    ]

    bloom_filter = None

    def __init__(
        self,
        header: BlockHeaderAPI,
        transactions: Sequence[SignedTransactionAPI] = None,
        uncles: Sequence[BlockHeaderAPI] = None,
        withdrawals: Sequence[WithdrawalAPI] = None,
    ) -> None:
        if transactions is None:
            transactions = []
        if uncles is None:
            uncles = []
        if withdrawals is None:
            withdrawals = []

        self.bloom_filter = BloomFilter(header.bloom)

        super().__init__(
            header=header,
            transactions=transactions,
            uncles=uncles,
            withdrawals=withdrawals,
        )

    @property
    def number(self) -> BlockNumber:
        return self.header.block_number

    @property
    def hash(self) -> Hash32:
        return self.header.hash

    @classmethod
    def get_transaction_builder(cls) -> Type[TransactionBuilderAPI]:
        return cls.transaction_builder

    @classmethod
    def get_receipt_builder(cls) -> Type[ReceiptBuilderAPI]:
        return cls.receipt_builder

    def get_receipts(self, chaindb: ChainDatabaseAPI) -> Tuple[ReceiptAPI, ...]:
        return chaindb.get_receipts(self.header, self.get_receipt_builder())

    @classmethod
    def from_header(
        cls,
        header: BlockHeaderAPI,
        chaindb: ChainDatabaseAPI,
    ) -> "ShanghaiBlock":
        """
        Returns the block denoted by the given block header.

        :raise eth.exceptions.BlockNotFound: if transactions, uncle headers,
               or withdrawals are missing
        """
        if header.uncles_hash == EMPTY_UNCLE_HASH:
            uncles: Tuple[BlockHeaderAPI, ...] = ()
        else:
            try:
                uncles = chaindb.get_block_uncles(header.uncles_hash)
            except HeaderNotFound as exc:
                raise BlockNotFound(
                    f"Uncles not found in database for {header}: {exc}"
                ) from exc

        try:
            transactions = chaindb.get_block_transactions(
                header, cls.get_transaction_builder()
            )
        except MissingTrieNode as exc:
            raise BlockNotFound(
                f"Transactions not found in database for {header}: {exc}"
            ) from exc

        try:
            withdrawals = chaindb.get_block_withdrawals(header)
        except MissingTrieNode as exc:
            raise BlockNotFound(
                f"Withdrawals not found in database for {header}: {exc}"
            ) from exc

        return cls(
            header=header,
            transactions=transactions,
            uncles=uncles,
            withdrawals=withdrawals,
        )



from .tangerine_whistle import (  # noqa: F401
    TangerineWhistleVM,
)
from .frontier import (  # noqa: F401
    FrontierVM,
)
from .homestead import (  # noqa: F401
    HomesteadVM,
)
from .spurious_dragon import (  # noqa: F401
    SpuriousDragonVM,
)
from .byzantium import (  # noqa: F401
    ByzantiumVM,
)
from .constantinople import (  # noqa: F401
    ConstantinopleVM,
)
from .petersburg import (  # noqa: F401
    PetersburgVM,
)
from .istanbul import (  # noqa: F401
    IstanbulVM,
)
from .muir_glacier import (  # noqa: F401
    MuirGlacierVM,
)
from .berlin import (  # noqa: F401
    BerlinVM,
)
from .london import (  # noqa: F401
    LondonVM,
)
from .arrow_glacier import (  # noqa: F401
    ArrowGlacierVM,
)
from .gray_glacier import (  # noqa: F401
    GrayGlacierVM,
)
from .paris import (  # noqa: F401
    ParisVM,
)
from .shanghai import (  # noqa: F401
    ShanghaiVM,
)



from eth.vm.forks.berlin.constants import (
    ACCESS_LIST_ADDRESS_COST_EIP_2930,
    ACCESS_LIST_STORAGE_KEY_COST_EIP_2930,
)

# EIP 1559
DYNAMIC_FEE_TRANSACTION_TYPE = 2
DYNAMIC_FEE_ADDRESS_COST = ACCESS_LIST_ADDRESS_COST_EIP_2930
DYNAMIC_FEE_STORAGE_KEY_COST = ACCESS_LIST_STORAGE_KEY_COST_EIP_2930

BASE_FEE_MAX_CHANGE_DENOMINATOR = 8
INITIAL_BASE_FEE = 1000000000
ELASTICITY_MULTIPLIER = 2

EIP3541_RESERVED_STARTING_BYTE = b"\xef"
EIP3529_MAX_REFUND_QUOTIENT = 5



import contextlib
import io
import logging
from typing import (
    Iterator,
    Set,
)

from eth.validation import (
    validate_is_bytes,
)
from eth.vm import (
    opcode_values,
)

PUSH1, PUSH32, STOP = opcode_values.PUSH1, opcode_values.PUSH32, opcode_values.STOP


class SlowCodeStream:
    """
    A known working version of code stream that is kept around for testing,
    despite not being optimized.
    """

    stream = None
    _length_cache = None
    _raw_code_bytes = None
    invalid_positions: Set[int] = None
    valid_positions: Set[int] = None

    logger = logging.getLogger("eth.vm.SlowCodeStream")

    def __init__(self, code_bytes: bytes) -> None:
        validate_is_bytes(code_bytes, title="SlowCodeStream bytes")
        stream = io.BytesIO(code_bytes)
        self.stream = stream
        self._bound_stream_read = stream.read
        self._raw_code_bytes = code_bytes
        self._length_cache = len(code_bytes)
        self.invalid_positions = set()
        self.valid_positions = set()

    def read(self, size: int) -> bytes:
        return self._bound_stream_read(size)

    def __len__(self) -> int:
        return self._length_cache

    def __getitem__(self, i: int) -> int:
        return self._raw_code_bytes[i]

    def __iter__(self) -> Iterator[int]:
        # a very performance-sensitive method
        read = self.read
        try:
            while True:
                yield ord(read(1))
        except TypeError:
            yield STOP

    def __next__(self) -> int:
        # a very performance-sensitive method
        next_opcode_as_byte = self._bound_stream_read(1)

        if next_opcode_as_byte:
            return ord(next_opcode_as_byte)
        else:
            return STOP

    def peek(self) -> int:
        current_pc = self.program_counter
        next_opcode = next(self)
        self.program_counter = current_pc
        return next_opcode

    @property
    def program_counter(self) -> int:
        return self.stream.tell()

    @program_counter.setter
    def program_counter(self, value: int) -> None:
        self.stream.seek(min(value, len(self)))

    @contextlib.contextmanager
    def seek(self, pc: int) -> Iterator["SlowCodeStream"]:
        anchor_pc = self.program_counter
        self.program_counter = pc
        try:
            yield self
        finally:
            self.program_counter = anchor_pc

    def _potentially_disqualifying_opcode_positions(
        self, position: int
    ) -> Iterator[int]:
        # Look at the last 32 positions (from 1 byte back to 32 bytes back).
        # Don't attempt to look at negative positions.
        deepest_lookback = min(32, position)
        # iterate in reverse, because PUSH32 is more common than others
        for bytes_back in range(deepest_lookback, 0, -1):
            earlier_position = position - bytes_back
            opcode = self._raw_code_bytes[earlier_position]
            if PUSH1 + (bytes_back - 1) <= opcode <= PUSH32:
                # that PUSH1, if two bytes back, isn't disqualifying
                # PUSH32 in any of the bytes back is disqualifying
                yield earlier_position

    def is_valid_opcode(self, position: int) -> bool:
        if position >= self._length_cache:
            return False
        elif position in self.invalid_positions:
            return False
        elif position in self.valid_positions:
            return True
        else:
            # An opcode is not valid, iff it is the "data" following a PUSH_
            # So we look at the previous 32 bytes (PUSH32 being the largest) to see if
            # there is a PUSH_ before the opcode in this position.
            for disqualifier in self._potentially_disqualifying_opcode_positions(
                position
            ):
                # Now that we found a PUSH_ before this position, we check if *that* PUSH is valid  # noqa: E501
                if self.is_valid_opcode(disqualifier):
                    # If the PUSH_ valid, then the current position is invalid
                    self.invalid_positions.add(position)
                    return False
                # Otherwise, keep looking for other potentially disqualifying PUSH_ codes  # noqa: E501

            # We didn't find any valid PUSH_ opcodes in the 32 bytes before position;it's valid  # noqa: E501
            self.valid_positions.add(position)
            return True



from collections.abc import (
    Mapping,
)
import itertools
from typing import (
    Any,
    Dict,
    Sequence,
)

from eth_utils.toolz import (
    merge_with,
)


def merge_if_dicts(values: Sequence[Dict[Any, Any]]) -> Any:
    if all(isinstance(item, Mapping) for item in values):
        return merge_with(merge_if_dicts, *values)
    else:
        return values[-1]


def deep_merge(*dicts: Dict[Any, Any]) -> Dict[Any, Any]:
    return merge_with(merge_if_dicts, *dicts)


def is_cleanly_mergable(*dicts: Dict[Any, Any]) -> bool:
    """
    Check that nothing will be overwritten when
    dictionaries are merged using `deep_merge`.

    Examples:

        >>> is_cleanly_mergable({"a": 1}, {"b": 2}, {"c": 3})
        True
        >>> is_cleanly_mergable({"a": 1}, {"b": 2}, {"a": 0, c": 3})
        False
        >>> is_cleanly_mergable({"a": 1, "b": {"ba": 2}}, {"c": 3, {"b": {"bb": 4}})
        True
        >>> is_cleanly_mergable({"a": 1, "b": {"ba": 2}}, {"b": {"ba": 4}})
        False

    """
    if len(dicts) <= 1:
        return True
    elif len(dicts) == 2:
        if not all(isinstance(d, Mapping) for d in dicts):
            return False
        else:
            shared_keys = set(dicts[0].keys()) & set(dicts[1].keys())
            return all(
                is_cleanly_mergable(dicts[0][key], dicts[1][key]) for key in shared_keys
            )
    else:
        dict_combinations = itertools.combinations(dicts, 2)
        return all(
            is_cleanly_mergable(*combination) for combination in dict_combinations
        )



import functools
import json
import os
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Tuple,
)

from eth_utils import (
    to_tuple,
)
from eth_utils.toolz import (
    curry,
    identity,
)

from ._utils import (
    recursive_find_files,
    require_pytest,
)


#
# Filesystem fixture loading.
#
def find_fixture_files(fixtures_base_dir: str) -> Iterable[str]:
    all_fixture_paths = recursive_find_files(fixtures_base_dir, "*.json")
    return all_fixture_paths


@to_tuple
def find_fixtures(fixtures_base_dir: str) -> Iterable[Tuple[str, str]]:
    """
    Finds all of the (fixture_path, fixture_key) pairs for a given path under
    the JSON test fixtures directory.
    """
    all_fixture_paths = find_fixture_files(fixtures_base_dir)

    for fixture_path in sorted(all_fixture_paths):
        with open(fixture_path) as fixture_file:
            fixtures = json.load(fixture_file)

        for fixture_key in sorted(fixtures.keys()):
            yield (fixture_path, fixture_key)


# we use an LRU cache on this function so that we can sort the tests such that
# all fixtures from the same file are executed sequentially allowing us to keep
# a small rolling cache of the loaded fixture files.
@functools.lru_cache(maxsize=16)
def load_json_fixture(fixture_path: str) -> Dict[str, Any]:
    """
    Loads a fixture file, caching the most recent files it loaded.
    """
    with open(fixture_path) as fixture_file:
        file_fixtures = json.load(fixture_file)
    return file_fixtures


def load_fixture(
    fixture_path: str, fixture_key: str, normalize_fn: Callable[..., Any] = identity
) -> Dict[str, Any]:
    """
    Loads a specific fixture from a fixture file, optionally passing it through
    a normalization function.
    """
    file_fixtures = load_json_fixture(fixture_path)
    fixture = normalize_fn(file_fixtures[fixture_key])
    return fixture


@require_pytest
@curry
def filter_fixtures(
    all_fixtures: Iterable[Any],
    fixtures_base_dir: str,
    mark_fn: Callable[[str, str], bool] = None,
    ignore_fn: Callable[..., bool] = None,
) -> Any:
    """
    Helper function for filtering test fixtures.

    - `fixtures_base_dir` should be the base dir that the fixtures were collected from.
    - `mark_fn` should be a func which either returns `None` or a `pytest.mark` object.
    - `ignore_fn` should be a function which returns `True` for any fixture
       which should be ignored.
    """
    import pytest  # noqa: F401

    for fixture_data in all_fixtures:
        fixture_path = fixture_data[0]
        fixture_relpath = os.path.relpath(fixture_path, fixtures_base_dir)

        if ignore_fn:
            if ignore_fn(fixture_relpath, *fixture_data[1:]):
                continue

        if mark_fn is not None:
            mark = mark_fn(fixture_relpath, *fixture_data[1:])
            if mark:
                yield pytest.param(
                    (fixture_path, *fixture_data[1:]),
                    marks=mark,
                )
                continue

        yield fixture_data



from typing import (
    Dict,
    Type,
)

from eth.rlp.receipts import (
    Receipt,
)
from eth.vm.forks.berlin.constants import (
    ACCESS_LIST_TRANSACTION_TYPE,
)
from eth.vm.forks.berlin.receipts import (
    BerlinReceiptBuilder,
    TypedReceipt as BerlinTypedReceipt,
)

from .constants import (
    DYNAMIC_FEE_TRANSACTION_TYPE,
)


class LondonTypedReceipt(BerlinTypedReceipt):
    codecs: Dict[int, Type[Receipt]] = {
        ACCESS_LIST_TRANSACTION_TYPE: Receipt,
        DYNAMIC_FEE_TRANSACTION_TYPE: Receipt,
    }


class LondonReceiptBuilder(BerlinReceiptBuilder):
    typed_receipt_class = LondonTypedReceipt



import struct
from typing import (
    Tuple,
)

doc = """
    Lovingly lifted from https://github.com/buggywhip/blake2_py
    with this license:

      Copyright (c) 2009-2018 Larry Bugbee, Kent, WA, USA

      Permission to use, copy, modify, and/or distribute this software
      for any purpose with or without fee is hereby granted, provided
      that the above copyright notice and this permission notice appear
      in all copies.

      THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL
      WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED
      WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
      THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR
      CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
      LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
      NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
      CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

        (the ISC license, a minor tweak of the BSD license)

"""


class Blake2:
    """Blake2 is a base class for Blake2b and Blake2s"""

    # for more than 10 rounds, the schedule wraps around to the beginning
    sigma_schedule = (
        (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15),
        (14, 10, 4, 8, 9, 15, 13, 6, 1, 12, 0, 2, 11, 7, 5, 3),
        (11, 8, 12, 0, 5, 2, 15, 13, 10, 14, 3, 6, 7, 1, 9, 4),
        (7, 9, 3, 1, 13, 12, 11, 14, 2, 6, 5, 10, 4, 0, 15, 8),
        (9, 0, 5, 7, 2, 4, 10, 15, 14, 1, 11, 12, 6, 8, 3, 13),
        (2, 12, 6, 10, 0, 11, 8, 3, 4, 13, 7, 5, 15, 14, 1, 9),
        (12, 5, 1, 15, 14, 13, 4, 10, 0, 7, 6, 3, 9, 2, 8, 11),
        (13, 11, 7, 14, 12, 1, 3, 9, 5, 0, 15, 4, 8, 6, 2, 10),
        (6, 15, 14, 9, 11, 3, 0, 8, 12, 2, 13, 7, 1, 4, 10, 5),
        (10, 2, 8, 4, 7, 6, 1, 5, 15, 11, 9, 14, 3, 12, 13, 0),
    )


class Blake2b(Blake2):
    WORDBITS = 64
    MASKBITS = 2**WORDBITS - 1
    WORDFMT = "Q"

    IV = (
        0x6A09E667F3BCC908,
        0xBB67AE8584CAA73B,
        0x3C6EF372FE94F82B,
        0xA54FF53A5F1D36F1,
        0x510E527FADE682D1,
        0x9B05688C2B3E6C1F,
        0x1F83D9ABFB41BD6B,
        0x5BE0CD19137E2179,
    )

    ROT1 = 32
    ROT2 = 24
    ROT3 = 16
    ROT4 = 63


TMessageBlock = Tuple[int, int, int, int, int, int, int, int]


def blake2b_compress(
    num_rounds: int,
    h_starting_state: TMessageBlock,
    block: bytes,
    t_offset_counters: Tuple[int, int],
    final_block_flag: bool,
) -> bytes:
    """
    'F Compression' from section 3.2 of RFC 7693:
    https://tools.ietf.org/html/rfc7693#section-3.2
    """

    # Dereference these for [very small] speed improvement.
    # Perhaps more than anything, this makes the code
    # easier to read.
    MASKBITS = Blake2b.MASKBITS
    WORDBITS = Blake2b.WORDBITS
    IV = Blake2b.IV
    ROT1 = Blake2b.ROT1
    ROT2 = Blake2b.ROT2
    ROT3 = Blake2b.ROT3
    ROT4 = Blake2b.ROT4
    WB_ROT1 = WORDBITS - ROT1
    WB_ROT2 = WORDBITS - ROT2
    WB_ROT3 = WORDBITS - ROT3
    WB_ROT4 = WORDBITS - ROT4

    sigma_schedule = Blake2b.sigma_schedule
    sigma_schedule_len = len(sigma_schedule)

    # convert block (bytes) into 16 LE words
    m = struct.unpack_from("<16%s" % Blake2b.WORDFMT, bytes(block))

    v = [0] * 16
    v[0:8] = h_starting_state
    v[8:12] = IV[:4]
    v[12] = t_offset_counters[0] ^ IV[4]
    v[13] = t_offset_counters[1] ^ IV[5]

    if final_block_flag:
        v[14] = Blake2b.MASKBITS ^ IV[6]
    else:
        v[14] = IV[6]

    # The original code had a mechanism to turn on a "tree mode",
    # setting f[1] to MASKBITS here.
    # There seems to be no reference to that bit flip in the 3.2 section of RFC 7693,
    # and there is no such setting in EIP-152. So the bit flip option is removed.
    v[15] = IV[7]

    # Within the confines of the Python language, this is a
    # highly optimized version of G().  It differs some from
    # the formal specification and reference implementation.
    def G(a: int, b: int, c: int, d: int) -> None:
        # dereference v[] for another small speed improvement
        va = v[a]
        vb = v[b]
        vc = v[c]
        vd = v[d]
        va = (va + vb + msri2) & MASKBITS
        w = vd ^ va
        vd = (w >> ROT1) | (w << (WB_ROT1)) & MASKBITS
        vc = (vc + vd) & MASKBITS
        w = vb ^ vc
        vb = (w >> ROT2) | (w << (WB_ROT2)) & MASKBITS
        va = (va + vb + msri21) & MASKBITS
        w = vd ^ va
        vd = (w >> ROT3) | (w << (WB_ROT3)) & MASKBITS
        vc = (vc + vd) & MASKBITS
        w = vb ^ vc
        vb = (w >> ROT4) | (w << (WB_ROT4)) & MASKBITS
        # re-reference v[]
        v[a] = va
        v[b] = vb
        v[c] = vc
        v[d] = vd

    # time to ChaCha
    for r in range(num_rounds):
        # resolve as much as possible outside G() and
        # don't pass as argument, let scope do its job.
        # Result is a 50% speed increase, but sadly,
        # "slow" divided by 1.5 is still "slow".  :-/
        sr = sigma_schedule[r % sigma_schedule_len]
        msri2 = m[sr[0]]
        msri21 = m[sr[1]]
        G(0, 4, 8, 12)
        msri2 = m[sr[2]]
        msri21 = m[sr[3]]
        G(1, 5, 9, 13)
        msri2 = m[sr[4]]
        msri21 = m[sr[5]]
        G(2, 6, 10, 14)
        msri2 = m[sr[6]]
        msri21 = m[sr[7]]
        G(3, 7, 11, 15)
        msri2 = m[sr[8]]
        msri21 = m[sr[9]]
        G(0, 5, 10, 15)
        msri2 = m[sr[10]]
        msri21 = m[sr[11]]
        G(1, 6, 11, 12)
        msri2 = m[sr[12]]
        msri21 = m[sr[13]]
        G(2, 7, 8, 13)
        msri2 = m[sr[14]]
        msri21 = m[sr[15]]
        G(3, 4, 9, 14)

    result_message_words = (h_starting_state[i] ^ v[i] ^ v[i + 8] for i in range(8))
    return struct.pack(f"<8{Blake2b.WORDFMT}", *result_message_words)



from eth.vm.forks.istanbul.computation import (
    ISTANBUL_PRECOMPILES,
    IstanbulComputation,
)

from .opcodes import (
    MUIR_GLACIER_OPCODES,
)

MUIR_GLACIER_PRECOMPILES = ISTANBUL_PRECOMPILES


class MuirGlacierComputation(IstanbulComputation):
    """
    A class for all execution *message* computations in the ``MuirGlacier`` fork.
    Inherits from
    :class:`~eth.vm.forks.constantinople.istanbul.IstanbulComputation`
    """

    # Override
    opcodes = MUIR_GLACIER_OPCODES
    _precompiles = MUIR_GLACIER_PRECOMPILES



from typing import (
    Callable,
)

from eth_utils import (
    ValidationError,
    get_extended_debug_logger,
)

from eth.abc import (
    GasMeterAPI,
)
from eth.exceptions import (
    OutOfGas,
)
from eth.validation import (
    validate_uint256,
)


def default_refund_strategy(gas_refunded_total: int, amount: int) -> int:
    if amount < 0:
        raise ValidationError("Gas refund amount must be positive")

    return gas_refunded_total + amount


def allow_negative_refund_strategy(gas_refunded_total: int, amount: int) -> int:
    return gas_refunded_total + amount


RefundStrategy = Callable[[int, int], int]


class GasMeter(GasMeterAPI):
    start_gas: int = None

    gas_refunded: int = None
    gas_remaining: int = None

    logger = get_extended_debug_logger("eth.gas.GasMeter")

    def __init__(
        self, start_gas: int, refund_strategy: RefundStrategy = default_refund_strategy
    ) -> None:
        validate_uint256(start_gas, title="Start Gas")

        self.refund_strategy = refund_strategy
        self.start_gas = start_gas

        self.gas_remaining = self.start_gas
        self.gas_refunded = 0

    #
    # Write API
    #
    def consume_gas(self, amount: int, reason: str) -> None:
        if amount < 0:
            raise ValidationError("Gas consumption amount must be positive")

        if amount > self.gas_remaining:
            raise OutOfGas(
                f"Out of gas: Needed {amount} "
                f"- Remaining {self.gas_remaining} "
                f"- Reason: {reason}"
            )

        self.gas_remaining -= amount

        if self.logger.show_debug2:
            self.logger.debug2(
                "GAS CONSUMPTION: %s - %s -> %s (%s)",
                self.gas_remaining + amount,
                amount,
                self.gas_remaining,
                reason,
            )

    def return_gas(self, amount: int) -> None:
        if amount < 0:
            raise ValidationError("Gas return amount must be positive")

        self.gas_remaining += amount

        if self.logger.show_debug2:
            self.logger.debug2(
                "GAS RETURNED: %s + %s -> %s",
                self.gas_remaining - amount,
                amount,
                self.gas_remaining,
            )

    def refund_gas(self, amount: int) -> None:
        self.gas_refunded = self.refund_strategy(self.gas_refunded, amount)

        if self.logger.show_debug2:
            self.logger.debug2(
                "GAS REFUND: %s + %s -> %s",
                self.gas_refunded - amount,
                amount,
                self.gas_refunded,
            )



from enum import (
    Enum,
)
from typing import (
    Dict,
    FrozenSet,
    List,
    NamedTuple,
    Sequence,
)

from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
)

from .constants import (
    NONCE_AUTH,
    NONCE_DROP,
)


class VoteAction(Enum):
    """
    The action that is being voted on (nominate or kick).
    """

    # Using the byte representation rather than auto()
    # makes serialization more convenient
    NOMINATE = NONCE_AUTH
    KICK = NONCE_DROP

    def validate_for(self, signers: Sequence[Address], subject: Address) -> None:
        """
        Check if a vote action is semantically valid, meaning it is either voting
        against an existing member or voting in favor of an address that is not
        yet a member.
        """
        signer_exists = subject in signers
        signer_is_kicked = signer_exists and self is VoteAction.KICK
        signer_is_nominated = not signer_exists and self is VoteAction.NOMINATE

        if not signer_is_kicked and not signer_is_nominated:
            raise ValidationError(
                "Must either kick an existing signer or nominate a new signer"
                f"Subject: {subject!r} Current signers: {signers}"
            )


class Tally(NamedTuple):
    """
    Represent a tally to track votes on new members getting signed in or out.
    """

    action: VoteAction
    votes: int = 0

    def upvote(self) -> "Tally":
        return Tally(self.action, self.votes + 1)

    def downvote(self) -> "Tally":
        return Tally(self.action, self.votes - 1)


class Vote(NamedTuple):
    """
    Represent a vote (nominate/kick) from a signer regarding a subject.
    """

    signer: Address
    block_number: int
    subject: Address
    action: VoteAction


class Snapshot(NamedTuple):
    """
    Represent the current state of the consensus at a given time. This
    includes all current signers, all current votes as well at all tallies.
    """

    signers: FrozenSet[Address]
    block_hash: Hash32
    votes: FrozenSet[Vote]
    # Unfortunately there's no FrozenDict but at least Address and Tally
    # are both immutable
    tallies: Dict[Address, Tally]

    def get_mutable_clone(self, block_hash: Hash32) -> "MutableSnapshot":
        """
        Return a ``MutableSnapshot`` snapshot clone that can be used as a work in
        progress representation of the snapshot.
        """
        return MutableSnapshot(
            signers=list(self.signers),
            block_hash=block_hash,
            votes=list(self.votes),
            tallies=self.tallies.copy(),
        )

    def get_sorted_signers(self) -> List[Address]:
        """
        Return the sorted list of signers.
        """
        return sorted(self.signers)


class MutableSnapshot(NamedTuple):
    """
    Like ``Snapshot`` but mutable as a work in progress representation.
    """

    signers: List[Address]
    block_hash: Hash32
    votes: List[Vote]
    tallies: Dict[Address, Tally]

    def get_immutable_clone(self) -> Snapshot:
        """
        Return a ``Snapshot``, an immutable clone of the mutable snapshot.
        """
        return Snapshot(
            signers=frozenset(self.signers),
            block_hash=self.block_hash,
            votes=frozenset(self.votes),
            tallies=self.tallies.copy(),
        )



from typing import (
    Dict,
    Sequence,
    Tuple,
    Type,
    Union,
)

from cached_property import (
    cached_property,
)
from eth_keys.datatypes import (
    PrivateKey,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    to_bytes,
)
import rlp
from rlp.sedes import (
    CountableList,
    big_endian_int,
    binary,
)

from eth._utils.transactions import (
    calculate_intrinsic_gas,
    create_transaction_signature,
    extract_transaction_sender,
    validate_transaction_signature,
)
from eth.abc import (
    ReceiptAPI,
    SignedTransactionAPI,
    TransactionDecoderAPI,
)
from eth.constants import (
    CREATE_CONTRACT_ADDRESS,
)
from eth.rlp.logs import (
    Log,
)
from eth.rlp.receipts import (
    Receipt,
)
from eth.rlp.sedes import (
    address,
)
from eth.rlp.transactions import (
    SignedTransactionMethods,
)
from eth.validation import (
    validate_canonical_address,
    validate_is_bytes,
    validate_is_transaction_access_list,
    validate_uint64,
    validate_uint256,
)
from eth.vm.forks.berlin.constants import (
    ACCESS_LIST_ADDRESS_COST_EIP_2930,
    ACCESS_LIST_STORAGE_KEY_COST_EIP_2930,
    ACCESS_LIST_TRANSACTION_TYPE,
)
from eth.vm.forks.berlin.transactions import (
    AccessListPayloadDecoder,
    AccountAccesses,
    BerlinLegacyTransaction,
    BerlinTransactionBuilder,
    BerlinUnsignedLegacyTransaction,
    TypedTransaction,
)
from eth.vm.forks.istanbul.transactions import (
    ISTANBUL_TX_GAS_SCHEDULE,
)

from .constants import (
    DYNAMIC_FEE_TRANSACTION_TYPE,
)
from .receipts import (
    LondonReceiptBuilder,
)


class LondonLegacyTransaction(BerlinLegacyTransaction):
    pass


class LondonUnsignedLegacyTransaction(BerlinUnsignedLegacyTransaction):
    def as_signed_transaction(
        self, private_key: PrivateKey, chain_id: int = None
    ) -> LondonLegacyTransaction:
        v, r, s = create_transaction_signature(self, private_key, chain_id=chain_id)
        return LondonLegacyTransaction(
            nonce=self.nonce,
            gas_price=self.gas_price,
            gas=self.gas,
            to=self.to,
            value=self.value,
            data=self.data,
            v=v,
            r=r,
            s=s,
        )


class UnsignedDynamicFeeTransaction(rlp.Serializable):
    _type_id = DYNAMIC_FEE_TRANSACTION_TYPE
    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("max_priority_fee_per_gas", big_endian_int),
        ("max_fee_per_gas", big_endian_int),
        ("gas", big_endian_int),
        ("to", address),
        ("value", big_endian_int),
        ("data", binary),
        ("access_list", CountableList(AccountAccesses)),
    ]

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self._type_id)

    def get_message_for_signing(self) -> bytes:
        payload = rlp.encode(self)
        return self._type_byte + payload

    def as_signed_transaction(self, private_key: PrivateKey) -> "TypedTransaction":
        message = self.get_message_for_signing()
        signature = private_key.sign_msg(message)
        y_parity, r, s = signature.vrs

        signed_transaction = DynamicFeeTransaction(
            self.chain_id,
            self.nonce,
            self.max_priority_fee_per_gas,
            self.max_fee_per_gas,
            self.gas,
            self.to,
            self.value,
            self.data,
            self.access_list,
            y_parity,
            r,
            s,
        )
        return LondonTypedTransaction(self._type_id, signed_transaction)

    def validate(self) -> None:
        validate_uint256(self.chain_id, title="Transaction.chain_id")
        validate_uint64(self.nonce, title="Transaction.nonce")
        validate_uint256(self.max_fee_per_gas, title="Transaction.max_fee_per_gas")
        validate_uint256(
            self.max_priority_fee_per_gas, title="Transaction.max_priority_fee_per_gas"
        )
        validate_uint256(self.gas, title="Transaction.gas")
        if self.to != CREATE_CONTRACT_ADDRESS:
            validate_canonical_address(self.to, title="Transaction.to")
        validate_uint256(self.value, title="Transaction.value")
        validate_is_bytes(self.data, title="Transaction.data")
        validate_is_transaction_access_list(self.access_list)

    def get_intrinsic_gas(self) -> int:
        return _get_dynamic_fee_txn_intrinsic_gas(self)

    @property
    def intrinsic_gas(self) -> int:
        return self.get_intrinsic_gas()


class DynamicFeeTransaction(
    rlp.Serializable, SignedTransactionMethods, SignedTransactionAPI
):
    _type_id = DYNAMIC_FEE_TRANSACTION_TYPE
    fields = [
        ("chain_id", big_endian_int),
        ("nonce", big_endian_int),
        ("max_priority_fee_per_gas", big_endian_int),
        ("max_fee_per_gas", big_endian_int),
        ("gas", big_endian_int),
        ("to", address),
        ("value", big_endian_int),
        ("data", binary),
        ("access_list", CountableList(AccountAccesses)),
        ("y_parity", big_endian_int),
        ("r", big_endian_int),
        ("s", big_endian_int),
    ]

    @property
    def gas_price(self) -> None:
        raise AttributeError(
            "Gas price is no longer available."
            "See max_priority_fee_per_gas or max_fee_per_gas"
        )

    def get_sender(self) -> Address:
        return extract_transaction_sender(self)

    def get_message_for_signing(self) -> bytes:
        unsigned = UnsignedDynamicFeeTransaction(
            self.chain_id,
            self.nonce,
            self.max_priority_fee_per_gas,
            self.max_fee_per_gas,
            self.gas,
            self.to,
            self.value,
            self.data,
            self.access_list,
        )
        payload = rlp.encode(unsigned)
        return self._type_byte + payload

    def check_signature_validity(self) -> None:
        validate_transaction_signature(self)

    @cached_property
    def _type_byte(self) -> bytes:
        return to_bytes(self._type_id)

    @cached_property
    def hash(self) -> Hash32:
        raise NotImplementedError("Call hash() on the TypedTransaction instead")

    def get_intrinsic_gas(self) -> int:
        return _get_dynamic_fee_txn_intrinsic_gas(self)

    def encode(self) -> bytes:
        return rlp.encode(self)

    def make_receipt(
        self,
        status: bytes,
        gas_used: int,
        log_entries: Tuple[Tuple[bytes, Tuple[int, ...], bytes], ...],
    ) -> ReceiptAPI:
        logs = [Log(address, topics, data) for address, topics, data in log_entries]
        # TypedTransaction is responsible for wrapping this into a TypedReceipt
        return Receipt(
            state_root=status,
            gas_used=gas_used,
            logs=logs,
        )


class DynamicFeePayloadDecoder(TransactionDecoderAPI):
    @classmethod
    def decode(cls, payload: bytes) -> SignedTransactionAPI:
        return rlp.decode(payload, sedes=DynamicFeeTransaction)


class LondonTypedTransaction(TypedTransaction):
    decoders: Dict[int, Type[TransactionDecoderAPI]] = {
        ACCESS_LIST_TRANSACTION_TYPE: AccessListPayloadDecoder,
        DYNAMIC_FEE_TRANSACTION_TYPE: DynamicFeePayloadDecoder,
    }
    receipt_builder = LondonReceiptBuilder


class LondonTransactionBuilder(BerlinTransactionBuilder):
    legacy_signed = LondonLegacyTransaction
    legacy_unsigned = LondonUnsignedLegacyTransaction
    typed_transaction = LondonTypedTransaction

    @classmethod
    def new_unsigned_dynamic_fee_transaction(
        cls,
        chain_id: int,
        nonce: int,
        max_priority_fee_per_gas: int,
        max_fee_per_gas: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        access_list: Sequence[Tuple[Address, Sequence[int]]],
    ) -> LondonTypedTransaction:
        transaction = UnsignedDynamicFeeTransaction(
            chain_id,
            nonce,
            max_priority_fee_per_gas,
            max_fee_per_gas,
            gas,
            to,
            value,
            data,
            access_list,
        )
        return transaction

    @classmethod
    def new_dynamic_fee_transaction(
        cls,
        chain_id: int,
        nonce: int,
        max_priority_fee_per_gas: int,
        max_fee_per_gas: int,
        gas: int,
        to: Address,
        value: int,
        data: bytes,
        access_list: Sequence[Tuple[Address, Sequence[int]]],
        y_parity: int,
        r: int,
        s: int,
    ) -> LondonTypedTransaction:
        transaction = DynamicFeeTransaction(
            chain_id,
            nonce,
            max_priority_fee_per_gas,
            max_fee_per_gas,
            gas,
            to,
            value,
            data,
            access_list,
            y_parity,
            r,
            s,
        )
        return LondonTypedTransaction(DYNAMIC_FEE_TRANSACTION_TYPE, transaction)


def _get_dynamic_fee_txn_intrinsic_gas(
    klass: Union[DynamicFeeTransaction, UnsignedDynamicFeeTransaction]
) -> int:
    core_gas = calculate_intrinsic_gas(ISTANBUL_TX_GAS_SCHEDULE, klass)

    num_addresses = len(klass.access_list)
    preload_address_costs = ACCESS_LIST_ADDRESS_COST_EIP_2930 * num_addresses

    num_slots = sum(len(slots) for _, slots in klass.access_list)
    preload_slot_costs = ACCESS_LIST_STORAGE_KEY_COST_EIP_2930 * num_slots

    return core_gas + preload_address_costs + preload_slot_costs



from typing import (
    Tuple,
    Type,
)

from eth_typing import BlockNumber
from eth_utils import (
    decode_hex,
    encode_hex,
    ValidationError,
)

from .constants import (
    MAINNET_CHAIN_ID,
    PARIS_MAINNET_BLOCK,
    GRAY_GLACIER_MAINNET_BLOCK,
    ARROW_GLACIER_MAINNET_BLOCK,
    LONDON_MAINNET_BLOCK,
    BERLIN_MAINNET_BLOCK,
    BYZANTIUM_MAINNET_BLOCK,
    PETERSBURG_MAINNET_BLOCK,
    ISTANBUL_MAINNET_BLOCK,
    MUIR_GLACIER_MAINNET_BLOCK,
    TANGERINE_WHISTLE_MAINNET_BLOCK,
    HOMESTEAD_MAINNET_BLOCK,
    SPURIOUS_DRAGON_MAINNET_BLOCK,
    DAO_FORK_MAINNET_EXTRA_DATA,
    DAO_FORK_MAINNET_BLOCK,
)
from eth import constants as eth_constants

from eth.abc import (
    BlockHeaderAPI,
    VirtualMachineAPI,
)
from eth.chains.base import (
    Chain,
)
from eth.rlp.headers import BlockHeader
from eth.vm.forks import (
    ArrowGlacierVM,
    BerlinVM,
    ByzantiumVM,
    FrontierVM,
    GrayGlacierVM,
    HomesteadVM,
    IstanbulVM,
    LondonVM,
    MuirGlacierVM,
    ParisVM,
    PetersburgVM,
    ShanghaiVM,
    SpuriousDragonVM,
    TangerineWhistleVM,
)


def validate_header_is_on_intended_dao_fork(
    support_dao_fork: bool, dao_fork_at: BlockNumber, header: BlockHeaderAPI
) -> None:
    # The special extra_data is set on the ten headers starting at the fork
    extra_data_block_nums = range(dao_fork_at, dao_fork_at + 10)

    if header.block_number in extra_data_block_nums:
        if support_dao_fork and header.extra_data != DAO_FORK_MAINNET_EXTRA_DATA:
            raise ValidationError(
                f"Block {header!r} must have extra data "
                f"{encode_hex(DAO_FORK_MAINNET_EXTRA_DATA)} not "
                f"{encode_hex(header.extra_data)} when supporting DAO fork"
            )
        elif not support_dao_fork and header.extra_data == DAO_FORK_MAINNET_EXTRA_DATA:
            raise ValidationError(
                f"Block {header!r} must not have extra data "
                f"{encode_hex(DAO_FORK_MAINNET_EXTRA_DATA)} when declining the DAO fork"
            )


class MainnetDAOValidatorVM(HomesteadVM):
    """
    Only on mainnet, TheDAO fork is accompanied by special extra data.
    Validate those headers
    """

    @classmethod
    def validate_header(
        cls, header: BlockHeaderAPI, previous_header: BlockHeaderAPI
    ) -> None:
        super().validate_header(header, previous_header)
        validate_header_is_on_intended_dao_fork(
            cls.support_dao_fork, cls.get_dao_fork_block_number(), header
        )


class MainnetHomesteadVM(MainnetDAOValidatorVM):
    _dao_fork_block_number = DAO_FORK_MAINNET_BLOCK


MAINNET_FORK_BLOCKS = (
    eth_constants.GENESIS_BLOCK_NUMBER,
    HOMESTEAD_MAINNET_BLOCK,
    TANGERINE_WHISTLE_MAINNET_BLOCK,
    SPURIOUS_DRAGON_MAINNET_BLOCK,
    BYZANTIUM_MAINNET_BLOCK,
    PETERSBURG_MAINNET_BLOCK,
    ISTANBUL_MAINNET_BLOCK,
    MUIR_GLACIER_MAINNET_BLOCK,
    BERLIN_MAINNET_BLOCK,
    LONDON_MAINNET_BLOCK,
    ARROW_GLACIER_MAINNET_BLOCK,
    GRAY_GLACIER_MAINNET_BLOCK,
    PARIS_MAINNET_BLOCK,
)
MINING_MAINNET_VMS = (
    FrontierVM,
    MainnetHomesteadVM,
    TangerineWhistleVM,
    SpuriousDragonVM,
    ByzantiumVM,
    PetersburgVM,
    IstanbulVM,
    MuirGlacierVM,
    BerlinVM,
    LondonVM,
    ArrowGlacierVM,
    GrayGlacierVM,
)
POS_MAINNET_VMS = (
    ParisVM,
    ShanghaiVM,
)

MAINNET_VMS = MINING_MAINNET_VMS + POS_MAINNET_VMS
MAINNET_VM_CONFIGURATION = tuple(zip(MAINNET_FORK_BLOCKS, MAINNET_VMS))


class BaseMainnetChain:
    chain_id = MAINNET_CHAIN_ID
    vm_configuration: Tuple[
        Tuple[BlockNumber, Type[VirtualMachineAPI]], ...
    ] = MAINNET_VM_CONFIGURATION


class MainnetChain(BaseMainnetChain, Chain):
    pass


MAINNET_GENESIS_HEADER = BlockHeader(
    difficulty=eth_constants.GENESIS_DIFFICULTY,
    extra_data=decode_hex(
        "0x11bbe8db4e347b4e8c937c1c8370e4b5ed33adb3db69cbdb7a38e1e50b1b82fa"
    ),
    gas_limit=eth_constants.GENESIS_GAS_LIMIT,
    gas_used=0,
    bloom=0,
    mix_hash=eth_constants.ZERO_HASH32,
    nonce=eth_constants.GENESIS_NONCE,
    block_number=0,
    parent_hash=eth_constants.ZERO_HASH32,
    receipt_root=eth_constants.BLANK_ROOT_HASH,
    uncles_hash=eth_constants.EMPTY_UNCLE_HASH,
    state_root=decode_hex(
        "0xd7f8974fb5ac78d9ac099b9ad5018bedc2ce0a72dad1827a1709da30580f0544"
    ),
    timestamp=0,
    transaction_root=eth_constants.BLANK_ROOT_HASH,
)



import copy
from typing import (
    Dict,
)

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.abc import (
    OpcodeAPI,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.london.opcodes import (
    LONDON_OPCODES,
)
from eth.vm.logic import (
    block,
)
from eth.vm.opcode import (
    as_opcode,
)

NEW_OPCODES = {
    # EIP-4399: supplant DIFFICULTY with PREVRANDAO
    opcode_values.PREVRANDAO: as_opcode(
        logic_fn=block.mixhash,
        mnemonic=mnemonics.PREVRANDAO,
        gas_cost=constants.GAS_BASE,
    ),
}

PARIS_OPCODES: Dict[int, OpcodeAPI] = merge(copy.deepcopy(LONDON_OPCODES), NEW_OPCODES)



from typing import (
    Type,
)

from eth.rlp.blocks import BaseBlock
from eth.vm.forks.istanbul import (
    IstanbulVM,
)
from eth.vm.state import BaseState

from .blocks import MuirGlacierBlock
from .headers import (
    compute_muir_glacier_difficulty,
    configure_muir_glacier_header,
    create_muir_glacier_header_from_parent,
)
from .state import MuirGlacierState


class MuirGlacierVM(IstanbulVM):
    # fork name
    fork = "muir-glacier"

    # classes
    block_class: Type[BaseBlock] = MuirGlacierBlock
    _state_class: Type[BaseState] = MuirGlacierState

    # Methods
    create_header_from_parent = staticmethod(create_muir_glacier_header_from_parent)  # type: ignore  # noqa: E501
    compute_difficulty = staticmethod(compute_muir_glacier_difficulty)  # type: ignore
    configure_header = configure_muir_glacier_header



from lru import (
    LRU,
)

from eth.abc import (
    DatabaseAPI,
)
from eth.db.backends.base import (
    BaseDB,
)


class CacheDB(BaseDB):
    """
    Set and get decoded RLP objects, where the underlying db stores
    encoded objects.
    """

    def __init__(self, db: DatabaseAPI, cache_size: int = 2048) -> None:
        self._db = db
        self._cache_size = cache_size
        self.reset_cache()

    def reset_cache(self) -> None:
        self._cached_values = LRU(self._cache_size)

    def __getitem__(self, key: bytes) -> bytes:
        if key not in self._cached_values:
            self._cached_values[key] = self._db[key]
        return self._cached_values[key]

    def __setitem__(self, key: bytes, value: bytes) -> None:
        self._cached_values[key] = value
        self._db[key] = value

    def __delitem__(self, key: bytes) -> None:
        if key in self._cached_values:
            del self._cached_values[key]
        del self._db[key]



import copy

from eth_utils.toolz import (
    merge,
)

from eth import (
    constants,
)
from eth.vm import (
    mnemonics,
    opcode_values,
)
from eth.vm.forks.byzantium.opcodes import (
    ensure_no_static,
)
from eth.vm.forks.istanbul.constants import (
    GAS_BALANCE_EIP1884,
    GAS_EXTCODEHASH_EIP1884,
    GAS_SLOAD_EIP1884,
)
from eth.vm.forks.petersburg.opcodes import (
    PETERSBURG_OPCODES,
)
from eth.vm.logic import (
    context,
    storage,
)
from eth.vm.opcode import (
    as_opcode,
)

from .storage import (
    sstore_eip2200,
)

UPDATED_OPCODES = {
    # New opcodes
    opcode_values.CHAINID: as_opcode(
        logic_fn=context.chain_id,
        mnemonic=mnemonics.CHAINID,
        gas_cost=constants.GAS_BASE,
    ),
    opcode_values.SELFBALANCE: as_opcode(
        logic_fn=context.selfbalance,
        mnemonic=mnemonics.SELFBALANCE,
        gas_cost=constants.GAS_LOW,
    ),
    # Repriced opcodes
    opcode_values.SSTORE: as_opcode(
        logic_fn=ensure_no_static(sstore_eip2200),
        mnemonic=mnemonics.SSTORE,
        gas_cost=constants.GAS_NULL,
    ),
    opcode_values.BALANCE: as_opcode(
        logic_fn=context.balance,
        mnemonic=mnemonics.BALANCE,
        gas_cost=GAS_BALANCE_EIP1884,
    ),
    opcode_values.SLOAD: as_opcode(
        logic_fn=storage.sload,
        mnemonic=mnemonics.SLOAD,
        gas_cost=GAS_SLOAD_EIP1884,
    ),
    opcode_values.EXTCODEHASH: as_opcode(
        logic_fn=context.extcodehash,
        mnemonic=mnemonics.EXTCODEHASH,
        gas_cost=GAS_EXTCODEHASH_EIP1884,
    ),
}

ISTANBUL_OPCODES = merge(
    copy.deepcopy(PETERSBURG_OPCODES),
    UPDATED_OPCODES,
)



from typing import (
    Sequence,
    Tuple,
    Type,
)

from eth_bloom import (
    BloomFilter,
)
from eth_typing import (
    BlockNumber,
    Hash32,
)
from rlp.sedes import (
    CountableList,
)
from trie.exceptions import (
    MissingTrieNode,
)

from eth.abc import (
    BlockHeaderAPI,
    ChainDatabaseAPI,
    ReceiptAPI,
    ReceiptBuilderAPI,
    SignedTransactionAPI,
    TransactionBuilderAPI,
)
from eth.constants import (
    EMPTY_UNCLE_HASH,
)
from eth.exceptions import (
    BlockNotFound,
    HeaderNotFound,
)
from eth.rlp.blocks import (
    BaseBlock,
)
from eth.rlp.headers import (
    BlockHeader,
)
from eth.rlp.receipts import (
    Receipt,
)

from .transactions import (
    FrontierTransaction,
)


class FrontierBlock(BaseBlock):
    transaction_builder = FrontierTransaction
    receipt_builder = Receipt
    fields = [
        ("header", BlockHeader),
        ("transactions", CountableList(transaction_builder)),
        ("uncles", CountableList(BlockHeader)),
    ]

    bloom_filter = None

    def __init__(
        self,
        header: BlockHeaderAPI,
        transactions: Sequence[SignedTransactionAPI] = None,
        uncles: Sequence[BlockHeaderAPI] = None,
    ) -> None:
        if transactions is None:
            transactions = []
        if uncles is None:
            uncles = []

        self.bloom_filter = BloomFilter(header.bloom)

        super().__init__(
            header=header,
            transactions=transactions,
            uncles=uncles,
        )
        # TODO: should perform block validation at this point?

    #
    # Helpers
    #
    @property
    def number(self) -> BlockNumber:
        return self.header.block_number

    @property
    def hash(self) -> Hash32:
        return self.header.hash

    #
    # Transaction class for this block class
    #
    @classmethod
    def get_transaction_builder(cls) -> Type[TransactionBuilderAPI]:
        return cls.transaction_builder

    @classmethod
    def get_receipt_builder(cls) -> Type[ReceiptBuilderAPI]:
        return cls.receipt_builder

    #
    # Receipts API
    #
    def get_receipts(self, chaindb: ChainDatabaseAPI) -> Tuple[ReceiptAPI, ...]:
        return chaindb.get_receipts(self.header, self.get_receipt_builder())

    #
    # Header API
    #
    @classmethod
    def from_header(
        cls, header: BlockHeaderAPI, chaindb: ChainDatabaseAPI
    ) -> "FrontierBlock":
        """
        Returns the block denoted by the given block header.

        :raise eth.exceptions.BlockNotFound: if transactions or uncle headers missing
        """
        if header.uncles_hash == EMPTY_UNCLE_HASH:
            uncles: Tuple[BlockHeaderAPI, ...] = ()
        else:
            try:
                uncles = chaindb.get_block_uncles(header.uncles_hash)
            except HeaderNotFound as exc:
                raise BlockNotFound(
                    f"Uncles not found in database for {header}: {exc}"
                ) from exc

        try:
            transactions = chaindb.get_block_transactions(
                header, cls.get_transaction_builder()
            )
        except MissingTrieNode as exc:
            raise BlockNotFound(
                f"Transactions not found in database for {header}: {exc}"
            ) from exc

        return cls(
            header=header,
            transactions=transactions,
            uncles=uncles,
        )



from typing import (
    Optional,
    Sequence,
    Tuple,
    cast,
)

from cached_property import (
    cached_property,
)
from eth_hash.auto import (
    keccak,
)
from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
)
import rlp
from rlp.sedes import (
    big_endian_int,
    binary,
)

from eth.abc import (
    BaseTransactionAPI,
    ComputationAPI,
    LegacyTransactionFieldsAPI,
    SignedTransactionAPI,
    TransactionBuilderAPI,
    TransactionFieldsAPI,
    UnsignedTransactionAPI,
)

from .sedes import (
    address,
)


class BaseTransactionMethods(BaseTransactionAPI):
    def validate(self) -> None:
        pass

    @property
    def intrinsic_gas(self) -> int:
        return self.get_intrinsic_gas()

    def gas_used_by(self, computation: ComputationAPI) -> int:
        return self.get_intrinsic_gas() + computation.get_gas_used()

    @property
    def chain_id(self) -> Optional[int]:
        return None

    @property
    def access_list(self) -> Sequence[Tuple[Address, Sequence[int]]]:
        return []


BASE_TRANSACTION_FIELDS = [
    ("nonce", big_endian_int),
    ("gas_price", big_endian_int),
    ("gas", big_endian_int),
    ("to", address),
    ("value", big_endian_int),
    ("data", binary),
    ("v", big_endian_int),
    ("r", big_endian_int),
    ("s", big_endian_int),
]


class BaseTransactionFields(rlp.Serializable, TransactionFieldsAPI):
    fields = BASE_TRANSACTION_FIELDS

    @property
    def hash(self) -> Hash32:
        return cast(Hash32, keccak(rlp.encode(self)))


class SignedTransactionMethods(BaseTransactionMethods, SignedTransactionAPI):
    type_id: Optional[int] = None

    @cached_property
    def sender(self) -> Address:
        return self.get_sender()

    #
    # Validation
    #
    def validate(self) -> None:
        if self.gas < self.intrinsic_gas:
            raise ValidationError("Insufficient gas")
        self.check_signature_validity()

    #
    # Signature and Sender
    #
    @property
    def is_signature_valid(self) -> bool:
        try:
            self.check_signature_validity()
        except ValidationError:
            return False
        else:
            return True


class BaseTransaction(
    LegacyTransactionFieldsAPI,
    BaseTransactionFields,
    SignedTransactionMethods,
    TransactionBuilderAPI,
):
    # "Legacy" transactions implemented by BaseTransaction are a combination of
    # the transaction codec (TransactionBuilderAPI) *and* the transaction
    # object (SignedTransactionAPI). In a multi-transaction-type world, that
    # becomes less desirable, and that responsibility splits up. See Berlin
    # transactions, for example.

    # Note that it includes at least one legacy field (v) that is not
    # explicitly accessible in new transaction types. See the v docstring in
    # LegacyTransactionFieldsAPI for more.

    # this is duplicated to make the rlp library happy, otherwise it complains
    # about no fields being defined but inheriting from multiple `Serializable`
    # bases.
    fields = BASE_TRANSACTION_FIELDS

    @classmethod
    def decode(cls, encoded: bytes) -> SignedTransactionAPI:
        return rlp.decode(encoded, sedes=cls)

    def encode(self) -> bytes:
        return rlp.encode(self)


class BaseUnsignedTransaction(
    BaseTransactionMethods, rlp.Serializable, UnsignedTransactionAPI
):
    fields = [
        ("nonce", big_endian_int),
        ("gas_price", big_endian_int),
        ("gas", big_endian_int),
        ("to", address),
        ("value", big_endian_int),
        ("data", binary),
    ]



import functools
from typing import (
    Any,
    Dict,
    Iterable,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from eth_typing import (
    Address,
    Hash32,
)
from eth_utils import (
    ValidationError,
    is_list_like,
)
from eth_utils.toolz import (
    dicttoolz,
    functoolz,
    itertoolz,
)

from eth._utils.headers import (
    compute_gas_limit_bounds,
)
from eth.abc import (
    VirtualMachineAPI,
)
from eth.constants import (
    SECPK1_N,
    UINT_64_MAX,
    UINT_256_MAX,
)
from eth.typing import (
    BytesOrView,
)


def validate_is_bytes(value: bytes, title: str = "Value") -> None:
    if not isinstance(value, bytes):
        raise ValidationError(f"{title} must be a byte string.  Got: {type(value)}")


def validate_is_bytes_or_view(value: BytesOrView, title: str = "Value") -> None:
    if isinstance(value, (bytes, memoryview)):
        return
    raise ValidationError(f"{title} must be bytes or memoryview. Got {type(value)}")


def validate_is_integer(value: Union[int, bool], title: str = "Value") -> None:
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValidationError(f"{title} must be a an integer.  Got: {type(value)}")


def validate_length(value: Sequence[Any], length: int, title: str = "Value") -> None:
    if not len(value) == length:
        raise ValidationError(
            f"{title} must be of length {length}.  Got {value} of length {len(value)}"
        )


def validate_length_lte(
    value: Sequence[Any], maximum_length: int, title: str = "Value"
) -> None:
    if len(value) > maximum_length:
        raise ValidationError(
            f"{title} must be of length less than or equal to {maximum_length}.  "
            f"Got {value} of length {len(value)}"
        )


def validate_gte(value: int, minimum: int, title: str = "Value") -> None:
    if value < minimum:
        raise ValidationError(
            f"{title} {value} is not greater than or equal to {minimum}"
        )
    validate_is_integer(value)


def validate_gt(value: int, minimum: int, title: str = "Value") -> None:
    if value <= minimum:
        raise ValidationError(f"{title} {value} is not greater than {minimum}")
    validate_is_integer(value, title=title)


def validate_lte(value: int, maximum: int, title: str = "Value") -> None:
    if value > maximum:
        raise ValidationError(f"{title} {value} is not less than or equal to {maximum}")
    validate_is_integer(value, title=title)


def validate_lt(value: int, maximum: int, title: str = "Value") -> None:
    if value >= maximum:
        raise ValidationError(f"{title} {value} is not less than {maximum}")
    validate_is_integer(value, title=title)


def validate_canonical_address(value: Address, title: str = "Value") -> None:
    if not isinstance(value, bytes) or not len(value) == 20:
        raise ValidationError(f"{title} {value!r} is not a valid canonical address")


def validate_multiple_of(value: int, multiple_of: int, title: str = "Value") -> None:
    if not value % multiple_of == 0:
        raise ValidationError(f"{title} {value} is not a multiple of {multiple_of}")


def validate_is_boolean(value: bool, title: str = "Value") -> None:
    if not isinstance(value, bool):
        raise ValidationError(f"{title} must be an boolean.  Got type: {type(value)}")


def validate_word(value: Hash32, title: str = "Value") -> None:
    if not isinstance(value, bytes):
        raise ValidationError(
            f"{title} is not a valid word. Must be of bytes type: Got: {type(value)}"
        )
    elif not len(value) == 32:
        raise ValidationError(
            f"{title} is not a valid word. Must be 32 bytes in length: "
            f"Got: {len(value)}"
        )


def validate_uint64(value: int, title: str = "Value") -> None:
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValidationError(f"{title} must be an integer: Got: {type(value)}")
    if value < 0:
        raise ValidationError(f"{title} cannot be negative: Got: {value}")
    if value > UINT_64_MAX:
        raise ValidationError(f"{title} exceeds maximum uint64 size.  Got: {value}")


def validate_uint256(value: int, title: str = "Value") -> None:
    if not isinstance(value, int) or isinstance(value, bool):
        raise ValidationError(f"{title} must be an integer: Got: {type(value)}")
    if value < 0:
        raise ValidationError(f"{title} cannot be negative: Got: {value}")
    if value > UINT_256_MAX:
        raise ValidationError(f"{title} exceeds maximum uint256 size.  Got: {value}")


def validate_stack_int(value: int) -> None:
    if 0 <= value <= UINT_256_MAX:
        return
    raise ValidationError(
        "Invalid Stack Item: Must be a 256 bit integer. Got {value!r}"
    )


def validate_stack_bytes(value: bytes) -> None:
    if len(value) <= 32:
        return
    raise ValidationError(
        "Invalid Stack Item: Must be either a length 32 byte string. Got {value!r}"
    )


validate_lt_secpk1n = functools.partial(validate_lte, maximum=SECPK1_N - 1)
validate_lt_secpk1n2 = functools.partial(validate_lte, maximum=SECPK1_N // 2 - 1)


def validate_unique(values: Iterable[Any], title: str = "Value") -> None:
    if not itertoolz.isdistinct(values):
        duplicates = functoolz.pipe(
            values,
            itertoolz.frequencies,  # get the frequencies
            # filter to ones that occure > 1
            functoolz.partial(dicttoolz.valfilter, lambda v: v > 1),
            sorted,  # sort them
            tuple,  # cast them to an immutiable form
        )
        raise ValidationError(
            f"{title} does not contain unique items.  Duplicates: "
            f"{', '.join((str(value) for value in duplicates))}"
        )


def validate_is_transaction_access_list(
    access_list: Sequence[Sequence[Union[Address, Sequence[int]]]]
) -> None:
    validate_is_list_like(access_list, "Transaction.access_list")
    for entry in access_list:
        validate_is_list_like(entry, "Transaction.access_list entry")
        if len(entry) != 2:
            raise ValidationError(
                "Transaction.access_list entry does not have 2 values. "
                "Needs address value and storage key list."
            )
        validate_canonical_address(
            cast(Address, entry[0]), "Transaction.access_list entry address"
        )
        validate_is_list_like(entry[1], "Transaction.access_list entry storage keys")
        for k in entry[1]:
            validate_is_integer(k, "Transaction.access_list entry storage key")


def validate_is_list_like(
    obj: Sequence[Any],
    title: str = "Value",
) -> None:
    if not is_list_like(obj):
        raise ValidationError(f"{title} is not list like: {repr(obj)}")


def validate_block_number(block_number: int, title: str = "Block Number") -> None:
    validate_is_integer(block_number, title)
    validate_gte(block_number, 0, title)


def validate_vm_block_numbers(vm_block_numbers: Iterable[int]) -> None:
    validate_unique(vm_block_numbers, title="Block Number set")

    for block_number in vm_block_numbers:
        validate_block_number(block_number)


def validate_vm_configuration(
    vm_configuration: Tuple[Tuple[int, Type[VirtualMachineAPI]], ...],
) -> None:
    validate_vm_block_numbers(
        tuple(block_number for block_number, _ in vm_configuration)
    )


def validate_gas_limit(gas_limit: int, parent_gas_limit: int) -> None:
    low_bound, high_bound = compute_gas_limit_bounds(parent_gas_limit)
    if gas_limit < low_bound:
        raise ValidationError(
            f"The gas limit {gas_limit} is too low. It must be at least {low_bound}"
        )
    elif gas_limit > high_bound:
        raise ValidationError(
            f"The gas limit {gas_limit} is too high. It must be at most {high_bound}"
        )


ALLOWED_HEADER_FIELDS = {
    "coinbase",
    "difficulty",
    "gas_limit",
    "timestamp",
    "extra_data",
    "mix_hash",
    "nonce",
    "uncles_hash",
    "transaction_root",
    "receipt_root",
    "withdrawals_root",
}


def validate_header_params_for_configuration(header_params: Dict[str, Any]) -> None:
    extra_fields = set(header_params.keys()).difference(ALLOWED_HEADER_FIELDS)
    if extra_fields:
        raise ValidationError(
            "The `configure_header` method may only be used with the fields "
            f"({', '.join(tuple(sorted(ALLOWED_HEADER_FIELDS)))}). "
            f"The provided fields ({', '.join(tuple(sorted(extra_fields)))}) "
            "are not supported"
        )



import itertools
import logging

from eth._utils.numeric import (
    ceil32,
)
from eth.abc import (
    MemoryAPI,
)
from eth.validation import (
    validate_is_bytes,
    validate_length,
    validate_lte,
    validate_uint256,
)


class Memory(MemoryAPI):
    __slots__ = ["_bytes"]
    logger = logging.getLogger("eth.vm.memory.Memory")

    def __init__(self) -> None:
        self._bytes = bytearray()

    def extend(self, start_position: int, size: int) -> None:
        if size == 0:
            return

        new_size = ceil32(start_position + size)
        if new_size <= len(self):
            return

        size_to_extend = new_size - len(self)
        try:
            self._bytes.extend(itertools.repeat(0, size_to_extend))
        except BufferError:
            # we can't extend the buffer (which might involve relocating it) if a
            # memoryview (which stores a pointer into the buffer) has been created by
            # read() and not released. Callers of read() will never try to write to the
            # buffer so we're not missing anything by making a new buffer and forgetting
            # about the old one. We're keeping too much memory around but this is still
            # a net savings over having read() return a new bytes() object every time.
            self._bytes = self._bytes + bytearray(size_to_extend)

    def __len__(self) -> int:
        return len(self._bytes)

    def write(self, start_position: int, size: int, value: bytes) -> None:
        if size:
            validate_uint256(start_position)
            validate_uint256(size)
            validate_is_bytes(value)
            validate_length(value, length=size)
            validate_lte(start_position + size, maximum=len(self))

            for idx, v in enumerate(value):
                self._bytes[start_position + idx] = v

    def read(self, start_position: int, size: int) -> memoryview:
        return memoryview(self._bytes)[start_position : start_position + size]

    def read_bytes(self, start_position: int, size: int) -> bytes:
        return bytes(self._bytes[start_position : start_position + size])



import os
from typing import (
    Any,
    Dict,
    Iterable,
    Tuple,
    Type,
    cast,
)

from eth_utils import (
    to_normalized_address,
)
from eth_utils.toolz import (
    assoc,
    first,
)
import rlp

from eth import (
    MainnetChain,
    constants,
)
from eth._utils.state import (
    diff_state,
)
from eth.abc import (
    BlockAPI,
    ChainAPI,
    StateAPI,
    VirtualMachineAPI,
)
from eth.chains.mainnet import (
    MainnetDAOValidatorVM,
)
from eth.db.atomic import (
    AtomicDB,
)
from eth.tools.builder.chain import (
    disable_pow_check,
)
from eth.typing import (
    AccountState,
)
from eth.vm.forks import (
    BerlinVM,
    ByzantiumVM,
    ConstantinopleVM,
    FrontierVM,
    GrayGlacierVM,
    HomesteadVM as BaseHomesteadVM,
    IstanbulVM,
    LondonVM,
    ParisVM,
    PetersburgVM,
    ShanghaiVM,
    SpuriousDragonVM,
    TangerineWhistleVM,
)


#
# State Setup
#
def setup_state(desired_state: AccountState, state: StateAPI) -> None:
    for account, account_data in desired_state.items():
        for slot, value in account_data["storage"].items():
            state.set_storage(account, slot, value)

        nonce = account_data["nonce"]
        code = account_data["code"]
        balance = account_data["balance"]

        state.set_nonce(account, nonce)
        state.set_code(account, code)
        state.set_balance(account, balance)
    state.persist()


def verify_state(expected_state: AccountState, state: StateAPI) -> None:
    diff = diff_state(expected_state, state)
    new_line = "\n"
    if diff:
        error_messages = []
        for account, field, actual_value, expected_value in diff:
            if field == "balance":
                error_messages.append(
                    f"{to_normalized_address(account)}(balance) | "
                    f"Actual: {actual_value!r} | Expected: {expected_value!r} | "
                    f"Delta: {cast(int, actual_value) - cast(int, expected_value)}"
                )
            else:
                error_messages.append(
                    f"{to_normalized_address(account)}({field}) | "
                    f"Actual: {actual_value!r} | Expected: {expected_value!r}"
                )
        raise AssertionError(
            f"State DB did not match expected state on {len(error_messages)} "
            f"values:{new_line} {f'{new_line} - '.join(error_messages)}"
        )


def chain_vm_configuration(
    fixture: Dict[str, Any]
) -> Iterable[Tuple[int, Type[VirtualMachineAPI]]]:
    network = fixture["network"]

    if network == "Frontier":
        return ((0, FrontierVM),)
    elif network == "Homestead":
        HomesteadVM = BaseHomesteadVM.configure(support_dao_fork=False)
        return ((0, HomesteadVM),)
    elif network == "EIP150":
        return ((0, TangerineWhistleVM),)
    elif network == "EIP158":
        return ((0, SpuriousDragonVM),)
    elif network == "Byzantium":
        return ((0, ByzantiumVM),)
    elif network == "Constantinople":
        return ((0, ConstantinopleVM),)
    elif network == "ConstantinopleFix":
        return ((0, PetersburgVM),)
    elif network == "Istanbul":
        return ((0, IstanbulVM),)
    elif network == "Berlin":
        return ((0, BerlinVM),)
    elif network == "London":
        return ((0, LondonVM),)
    elif network == "Merge":
        return ((0, ParisVM),)
    elif network == "Shanghai":
        return ((0, ShanghaiVM),)
    elif network == "FrontierToHomesteadAt5":
        HomesteadVM = BaseHomesteadVM.configure(support_dao_fork=False)
        return (
            (0, FrontierVM),
            (5, HomesteadVM),
        )
    elif network == "HomesteadToEIP150At5":
        HomesteadVM = BaseHomesteadVM.configure(support_dao_fork=False)
        return (
            (0, HomesteadVM),
            (5, TangerineWhistleVM),
        )
    elif network == "HomesteadToDaoAt5":
        HomesteadVM = MainnetDAOValidatorVM.configure(
            support_dao_fork=True,
            _dao_fork_block_number=5,
        )
        return ((0, HomesteadVM),)
    elif network == "EIP158ToByzantiumAt5":
        return (
            (0, SpuriousDragonVM),
            (5, ByzantiumVM),
        )
    elif network == "ByzantiumToConstantinopleFixAt5":
        return (
            (0, ByzantiumVM),
            (5, PetersburgVM),
        )
    elif network == "BerlinToLondonAt5":
        return (
            (0, BerlinVM),
            (5, LondonVM),
        )
    elif network == "ArrowGlacierToMergeAtDiffC0000":
        # Transition expected at 6 for all tests written thus far
        return (
            # These tests were written before Gray Glacier was a thing. Use
            # GrayGlacierVM instead since that's when PoW -> PoS transition happens.
            # They should both pass but this is the more logical transition to test.
            (0, GrayGlacierVM),
            (6, ParisVM),
        )
    elif network == "MergeToShanghaiAtTime15k":
        # Transition expected at 5 (timestamp==15000) for all tests written thus far
        return (
            (0, ParisVM),
            (5, ShanghaiVM),
        )

    else:
        raise ValueError(f"Network {network} does not match any known VM rules")


def genesis_fields_from_fixture(fixture: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert all genesis fields in a fixture to a dictionary of header fields and values.
    """

    header_fields = fixture["genesisBlockHeader"]
    base_fields = {
        "parent_hash": header_fields["parentHash"],
        "uncles_hash": header_fields["uncleHash"],
        "coinbase": header_fields["coinbase"],
        "state_root": header_fields["stateRoot"],
        "transaction_root": header_fields["transactionsTrie"],
        "receipt_root": header_fields["receiptTrie"],
        "bloom": header_fields["bloom"],
        "difficulty": header_fields["difficulty"],
        "block_number": header_fields["number"],
        "gas_limit": header_fields["gasLimit"],
        "gas_used": header_fields["gasUsed"],
        "timestamp": header_fields["timestamp"],
        "extra_data": header_fields["extraData"],
        "mix_hash": header_fields["mixHash"],
        "nonce": header_fields["nonce"],
    }
    if "baseFeePerGas" in header_fields:
        return assoc(base_fields, "base_fee_per_gas", header_fields["baseFeePerGas"])
    else:
        return base_fields


def genesis_params_from_fixture(fixture: Dict[str, Any]) -> Dict[str, Any]:
    """
    Convert a genesis fixture into a dict of the configurable header fields and values.

    Some fields cannot be explicitly set when creating a new header, like
    parent_hash, which is automatically set to the empty hash.
    """

    params = genesis_fields_from_fixture(fixture)

    # Confirm that (currently) non-configurable defaults are set correctly,
    #   then remove them because they cannot be configured on the header.
    defaults = (
        ("parent_hash", constants.GENESIS_PARENT_HASH),
        ("uncles_hash", constants.EMPTY_UNCLE_HASH),
        ("bloom", constants.GENESIS_BLOOM),
        ("block_number", constants.GENESIS_BLOCK_NUMBER),
        ("gas_used", constants.GENESIS_GAS_USED),
    )

    for key, default_val in defaults:
        supplied_val = params.pop(key)
        if supplied_val != default_val:
            raise ValueError(
                f"Unexpected genesis {key}: {supplied_val}, expected: {default_val}"
            )

    return params


def new_chain_from_fixture(
    fixture: Dict[str, Any], chain_cls: Type[ChainAPI] = MainnetChain
) -> ChainAPI:
    base_db = AtomicDB()

    vm_config = chain_vm_configuration(fixture)

    ChainFromFixture = chain_cls.configure(
        "ChainFromFixture",
        vm_configuration=vm_config,
    )

    if "sealEngine" in fixture and fixture["sealEngine"] == "NoProof":
        ChainFromFixture = disable_pow_check(ChainFromFixture)

    return ChainFromFixture.from_genesis(
        base_db,
        genesis_params=genesis_params_from_fixture(fixture),
        genesis_state=fixture["pre"],
    )


def apply_fixture_block_to_chain(
    block_fixture: Dict[str, Any], chain: ChainAPI, perform_validation: bool = True
) -> Tuple[BlockAPI, BlockAPI, bytes]:
    """
    :return: (provided_block, imported_block, rlp_encoded_imported_block)
    """
    # The block to import may be in a different block-class-range than the
    # chain's current one, so we use the block number specified in the
    # fixture to look up the correct block class.
    if "blockHeader" in block_fixture:
        block_number = block_fixture["blockHeader"]["number"]
        block_class = chain.get_vm_class_for_block_number(
            block_number
        ).get_block_class()
    else:
        block_class = chain.get_vm().get_block_class()

    block = rlp.decode(block_fixture["rlp"], sedes=block_class)

    import_result = chain.import_block(block, perform_validation=perform_validation)
    imported_block = import_result.imported_block

    rlp_encoded_imported_block = rlp.encode(imported_block, sedes=block_class)

    return (block, imported_block, rlp_encoded_imported_block)


def should_run_slow_tests() -> bool:
    if os.environ.get("TRAVIS_EVENT_TYPE") == "cron":
        return True
    return False


def get_test_name(filler: Dict[str, Any]) -> str:
    assert len(filler) == 1
    return first(filler)



from typing import (
    Type,
)

from eth.abc import (
    TransactionExecutorAPI,
    WithdrawalAPI,
)

from ..paris import (
    ParisState,
)
from ..paris.state import (
    ParisTransactionExecutor,
)
from .computation import (
    ShanghaiComputation,
)


class ShanghaiTransactionExecutor(ParisTransactionExecutor):
    pass


class ShanghaiState(ParisState):
    computation_class = ShanghaiComputation
    transaction_executor_class: Type[
        TransactionExecutorAPI
    ] = ShanghaiTransactionExecutor

    def apply_withdrawal(self, withdrawal: WithdrawalAPI) -> None:
        # withdrawal amount is in gwei, convert to wei
        amount_in_wei = withdrawal.amount * 10**9
        self.delta_balance(withdrawal.address, amount_in_wei)
